import datetime

from .arn import ARN
from .base_control import (
    Describer,
    MultiLister,
    NoResults,
    ResourceLister,
    format_timedelta,
)
from .common import Common


class CFNResourceLister(ResourceLister):
    prefix = "cfn_list"
    title = "CloudFormation Stacks"
    command_palette = ["cfn", "cloudformation"]

    def __init__(self, *args, **kwargs):
        self.resource_key = "cloudformation"
        self.list_method = "describe_stacks"
        self.item_path = ".Stacks"
        self.column_paths = {
            "name": ".StackName",
            "status": ".StackStatus",
            "drift": ".DriftInformation.StackDriftStatus",
            "created": self.determine_created,
            "updated": self.determine_updated,
        }
        self.hidden_columns = {
            "arn": ".StackId",
        }
        self.imported_column_sizes = {
            "name": 30,
            "status": 15,
            "drift": 15,
            "created": 20,
            "updated": 20,
        }
        self.describe_command = CFNDescriber.opener
        self.open_command = CFNRelated.opener
        self.open_selection_arg = "compare_value"

        self.imported_column_order = ["name", "status", "drift", "created", "updated"]
        self.sort_column = "name"
        self.primary_key = "name"
        super().__init__(*args, **kwargs)
        self.add_hotkey("p", self.parameters, "Parameters")
        self.add_hotkey("o", self.outputs, "Outputs")

    def parameters(self, _):
        if self.selection is not None:
            Common.Session.push_frame(CFNParameters.opener(stack=self.selection))

    def outputs(self, _):
        if self.selection is not None:
            Common.Session.push_frame(CFNOutputs.opener(stack=self.selection))

    def determine_created(self, cfn):
        return format_timedelta(
            datetime.datetime.now(datetime.timezone.utc)
            - datetime.datetime.fromisoformat(cfn["CreationTime"])
        )

    def determine_updated(self, cfn):
        if "LastUpdatedTime" in cfn:
            return format_timedelta(
                datetime.datetime.now(datetime.timezone.utc)
                - datetime.datetime.fromisoformat(cfn["LastUpdatedTime"])
            )
        return self.determine_created(cfn)


class CFNDescriber(Describer):
    prefix = "cfn_browser"
    title = "CloudFormation Stack"

    def populate_entry(self, *args, **kwargs):
        super().populate_entry(*args, **kwargs)
        try:
            arn = ARN(self.entry_id)
            self.entry_id = arn.resource_id_first
        except ValueError:
            return

    def __init__(
        self, parent, alignment, dimensions, entry, *args, entry_key="name", **kwargs
    ):
        self.resource_key = "cloudformation"
        self.describe_method = "describe_stacks"
        self.describe_kwarg_name = "StackName"
        self.object_path = ".Stacks[0]"
        super().__init__(
            parent,
            alignment,
            dimensions,
            *args,
            entry=entry,
            entry_key=entry_key,
            **kwargs
        )


class CFNRelated(MultiLister):
    prefix = "cfn_related"
    title = "Resources in CloudFormation Stack"

    def title_info(self):
        return self.compare_value

    def __init__(self, *args, **kwargs):
        kwargs["compare_key"] = "arn"
        self.resource_descriptors = [
            {
                "resource_key": "ec2",
                "list_method": "describe_instances",
                "list_kwargs": {},
                "item_path": "[.Reservations[].Instances[]]",
                "column_paths": {
                    "type": lambda x: "EC2 Instance",
                    "id": ".InstanceId",
                    "name": self.tag_finder_generator("Name"),
                },
                "hidden_columns": {},
                "compare_as_list": False,
                "compare_path": '.Tags[] | select(.Key=="aws:cloudformation:stack-id").Value',
            },
            {
                "resource_key": "autoscaling",
                "list_method": "describe_auto_scaling_groups",
                "list_kwargs": {},
                "item_path": ".AutoScalingGroups",
                "column_paths": {
                    "type": lambda x: "Autoscaling Group",
                    "id": self.empty,
                    "name": ".AutoScalingGroupName",
                },
                "hidden_columns": {},
                "compare_as_list": False,
                "compare_path": '.Tags[] | select(.Key=="aws:cloudformation:stack-id").Value',
            },
            {
                "resource_key": "rds",
                "list_method": "describe_db_instances",
                "list_kwargs": {},
                "item_path": ".DBInstances",
                "column_paths": {
                    "type": lambda x: "RDS Instance",
                    "id": ".DBInstanceIdentifier",
                    "name": ".Endpoint.Address",
                },
                "hidden_columns": {},
                "compare_as_list": False,
                "compare_path": '.TagList[] | select(.Key=="aws:cloudformation:stack-id").Value',
            },
            {
                "resource_key": "elbv2",
                "list_method": "describe_load_balancers",
                "list_kwargs": {},
                "item_path": ".LoadBalancers",
                "column_paths": {
                    "type": lambda x: "Load Balancer",
                    "id": self.empty,
                    "name": ".LoadBalancerName",
                },
                "hidden_columns": {
                    "arn": ".LoadBalancerArn",
                },
                "compare_as_list": False,
                "compare_path": self.comparer_generator(
                    "AWS::ElasticLoadBalancingV2::LoadBalancer", ".LoadBalancerArn"
                ),
            },
            {
                "resource_key": "elbv2",
                "list_method": "describe_target_groups",
                "list_kwargs": {},
                "item_path": ".TargetGroups",
                "column_paths": {
                    "type": lambda x: "Target Group",
                    "id": self.resource_id_from_arn_generator(".TargetGroupArn"),
                    "name": ".TargetGroupName",
                },
                "hidden_columns": {
                    "arn": ".TargetGroupArn",
                },
                "compare_as_list": False,
                "compare_path": self.comparer_generator(
                    "AWS::ElasticLoadBalancingV2::TargetGroup", ".TargetGroupArn"
                ),
            },
            {
                "resource_key": "elbv2",
                "list_method": "describe_listeners",
                "list_kwargs": self.kwargs_from_physids_generator(
                    "ListenerArns", "AWS::ElasticLoadBalancingV2::Listener"
                ),
                "item_path": ".Listeners",
                "column_paths": {
                    "type": lambda x: "Listener",
                    "id": self.full_resource_id_from_arn_generator(".ListenerArn"),
                    "name": self.empty,
                },
                "hidden_columns": {
                    "arn": ".ListenerArn",
                },
                "compare_as_list": False,
                "compare_path": self.comparer_generator(
                    "AWS::ElasticLoadBalancingV2::Listener", ".ListenerArn"
                ),
            },
            {
                "resource_key": "ec2",
                "list_method": "describe_vpcs",
                "list_kwargs": {},
                "item_path": ".Vpcs",
                "column_paths": {
                    "type": lambda x: "VPC",
                    "id": ".VpcId",
                    "name": self.tag_finder_generator("Name"),
                },
                "hidden_columns": {},
                "compare_as_list": False,
                "compare_path": '.Tags[] | select(.Key=="aws:cloudformation:stack-id").Value',
            },
            {
                "resource_key": "ec2",
                "list_method": "describe_subnets",
                "list_kwargs": {},
                "item_path": ".Subnets",
                "column_paths": {
                    "type": lambda x: "VPC Subnet",
                    "id": ".SubnetId",
                    "name": self.tag_finder_generator("Name"),
                },
                "hidden_columns": {},
                "compare_as_list": False,
                "compare_path": '.Tags[] | select(.Key=="aws:cloudformation:stack-id").Value',
            },
            {
                "resource_key": "route53",
                "list_method": "list_hosted_zones",
                "list_kwargs": {},
                "item_path": ".HostedZones",
                "column_paths": {
                    "type": lambda x: "Route53 Zone",
                    "id": ".Id",
                    "name": ".Name",
                },
                "hidden_columns": {},
                "compare_as_list": False,
                "compare_path": self.comparer_generator(
                    "AWS::Route53::HostedZone", ".Id"
                ),
            },
            {
                "resource_key": "ec2",
                "list_method": "describe_vpcs",
                "list_kwargs": {},
                "item_path": ".Vpcs",
                "column_paths": {
                    "type": lambda x: "VPC",
                    "id": ".VpcId",
                    "name": self.tag_finder_generator("Name"),
                },
                "hidden_columns": {},
                "compare_as_list": False,
                "compare_path": '.Tags[] | select(.Key=="aws:cloudformation:stack-id").Value',
            },
            {
                "resource_key": "ec2",
                "list_method": "describe_subnets",
                "list_kwargs": {},
                "item_path": ".Subnets",
                "column_paths": {
                    "type": lambda x: "Subnet",
                    "id": ".SubnetId",
                    "name": self.tag_finder_generator("Name"),
                },
                "hidden_columns": {},
                "compare_as_list": False,
                "compare_path": '.Tags[] | select(.Key=="aws:cloudformation:stack-id").Value',
            },
            {
                "resource_key": "ec2",
                "list_method": "describe_route_tables",
                "list_kwargs": {},
                "item_path": ".RouteTables",
                "column_paths": {
                    "type": lambda x: "Route Table",
                    "id": ".RouteTableId",
                    "name": self.tag_finder_generator("Name"),
                },
                "hidden_columns": {},
                "compare_as_list": False,
                "compare_path": '.Tags[] | select(.Key=="aws:cloudformation:stack-id").Value',
            },
        ]
        super().__init__(*args, **kwargs)

    def full_resource_id_from_arn_generator(self, arn_path):
        def fn(raw_item):
            arn = ARN(Common.Session.jq(arn_path).input(raw_item).first())
            return arn.resource_id

        return fn

    def resource_id_from_arn_generator(self, arn_path):
        def fn(raw_item):
            arn = ARN(Common.Session.jq(arn_path).input(raw_item).first())
            return arn.resource_id_first

        return fn

    def comparer_generator(self, cfn_type, physical_id_path):
        def fn(raw_item):
            phys_id = Common.Session.jq(physical_id_path).input(raw_item).first()
            if (
                cfn_type in self.stack_res_list
                and phys_id in self.stack_res_list[cfn_type]
            ):
                return self.compare_value
            return None

        return fn

    def kwargs_from_physids_generator(self, kwarg, cfn_type):
        def fn():
            if cfn_type in self.stack_res_list:
                return {kwarg: self.stack_res_list[cfn_type]}
            raise NoResults

        return fn

    def async_inner(self, *args, fn, clear=False, **kwargs):
        stop = False
        rl = None
        self.stack_res_list = {}
        while not stop:
            rargs = {"StackName": self.orig_compare_value["name"]}
            if rl is not None and "NextToken" in rl and rl["NextToken"] is not None:
                rargs["NextToken"] = rl["NextToken"]
            rl = Common.Session.service_provider("cloudformation").list_stack_resources(
                **rargs
            )
            if "NextToken" not in rl or rl["NextToken"] is None:
                stop = True
            for item in rl["StackResourceSummaries"]:
                if item["ResourceType"] not in self.stack_res_list:
                    self.stack_res_list[item["ResourceType"]] = []
                self.stack_res_list[item["ResourceType"]].append(
                    item["PhysicalResourceId"]
                )
        return super().async_inner(*args, fn=fn, clear=clear, **kwargs)


class CFNParameters(ResourceLister):
    prefix = "cfn_list"
    title = "CloudFormation Stack Parameters"

    def title_info(self):
        return self.stack_name

    def __init__(self, *args, stack=None, **kwargs):
        self.resource_key = "cloudformation"
        self.list_method = "describe_stacks"
        self.item_path = ".Stacks[0].Parameters"
        self.column_paths = {
            "name": ".ParameterKey",
            "value": self.determine_value,
        }
        self.list_kwargs = {"StackName": stack["name"]}
        self.imported_column_sizes = {
            "name": 30,
            "value": 30,
        }
        self.imported_column_order = ["name", "value"]

        self.sort_column = "name"
        self.primary_key = "name"
        self.stack_name = stack["name"]
        super().__init__(*args, **kwargs)

    def determine_value(self, param):
        if param["ParameterValue"] == "*" * len(param["ParameterValue"]):
            return "<hidden>"
        return param["ParameterValue"]


class CFNOutputs(ResourceLister):
    prefix = "cfn_list"
    title = "CloudFormation Stack Outputs"

    def title_info(self):
        return self.stack_name

    def __init__(self, *args, stack=None, **kwargs):
        self.resource_key = "cloudformation"
        self.list_method = "describe_stacks"
        self.item_path = ".Stacks[0].Outputs"
        self.column_paths = {
            "name": ".OutputKey",
            "value": ".OutputValue",
        }
        self.list_kwargs = {"StackName": stack["name"]}
        self.imported_column_sizes = {
            "name": 30,
            "value": 30,
        }
        self.imported_column_order = ["name", "value"]

        self.sort_column = "name"
        self.primary_key = "name"
        self.stack_name = stack["name"]
        super().__init__(*args, **kwargs)



from .base_control import Describer, ResourceLister
from .common import Common, SessionAwareDialog
from .termui.alignment import CenterAnchor, Dimension
from .termui.control import Border
from .termui.dialog import DialogFieldCheckbox, DialogFieldLabel, DialogFieldText
from .termui.ui import ControlCodes


class ASGResourceLister(ResourceLister):
    prefix = "asg_list"
    title = "Autoscaling Groups"
    command_palette = ["asg", "autoscaling"]

    def title_info(self):
        return self.title_info_data

    def matches(self, list_entry, *args):
        if self.lc is not None:
            if list_entry["launch config"] != self.lc["name"]:
                return False
        return super().matches(list_entry, *args)

    def __init__(self, *args, **kwargs):
        from .resource_ec2 import EC2ResourceLister

        self.resource_key = "autoscaling"
        self.list_method = "describe_auto_scaling_groups"
        self.title_info_data = None
        self.lc = None
        if "lc" in kwargs:
            lc = kwargs["lc"]
            self.title_info_data = "LaunchConfiguration: {0}".format(lc["name"])
            self.lc = lc

        self.item_path = ".AutoScalingGroups"
        self.column_paths = {
            "name": ".AutoScalingGroupName",
            "launch config/template": self.determine_launch_info,
            "current": self.determine_instance_count,
            "min": ".MinSize",
            "desired": ".DesiredCapacity",
            "max": ".MaxSize",
        }
        self.hidden_columns = {
            "launch config": ".LaunchConfigurationName",
            "arn": ".AutoScalingGroupARN",
        }
        self.imported_column_sizes = {
            "name": 30,
            "launch config/template": 30,
            "current": 10,
            "min": 10,
            "desired": 10,
            "max": 10,
        }
        self.describe_command = ASGDescriber.opener
        self.open_command = EC2ResourceLister.opener
        self.open_selection_arg = "asg"
        self.imported_column_order = [
            "name",
            "launch config/template",
            "current",
            "min",
            "desired",
            "max",
        ]

        self.sort_column = "name"
        self.primary_key = "name"
        super().__init__(*args, **kwargs)
        self.add_hotkey(ControlCodes.S, self.scale_group, "Scale")

    def determine_launch_info(self, asg):
        if "LaunchConfigurationName" in asg and bool(asg["LaunchConfigurationName"]):
            return asg["LaunchConfigurationName"]
        elif "LaunchTemplate" in asg:
            return asg["LaunchTemplate"]["LaunchTemplateName"]
        else:
            return ""

    def determine_instance_count(self, asg):
        return "{0}/{1}".format(
            len([h for h in asg["Instances"] if h["HealthStatus"] == "Healthy"]),
            len(asg["Instances"]),
        )

    def scale_group(self, _):
        if self.selection is not None:
            ASGScaleDialog(
                self.parent,
                CenterAnchor(0, 0),
                Dimension("80%|40", "20"),
                caller=self,
                weight=-500,
            )


class ASGScaleDialog(SessionAwareDialog):
    def __init__(self, *args, caller=None, **kwargs):
        kwargs["border"] = Border(
            Common.border("default"),
            Common.color("modal_dialog_border"),
            "Scale autoscaling group",
            Common.color("modal_dialog_border_title"),
            caller.selection["name"],
            Common.color("modal_dialog_border_title_info"),
        )
        self.asg_entry = caller.selection
        super().__init__(caller=caller, *args, **kwargs)
        self.error_label = DialogFieldLabel(
            "", default_color=Common.color("modal_dialog_error")
        )
        self.add_field(self.error_label)
        self.desired_capacity_field = DialogFieldText(
            "Desired capacity:",
            text=str(self.asg_entry["desired"]),
            label_min=16,
            color=Common.color("modal_dialog_textfield"),
            selected_color=Common.color("modal_dialog_textfield_selected"),
            label_color=Common.color("modal_dialog_textfield_label"),
            accepted_inputs="0123456789",
        )
        self.add_field(self.desired_capacity_field)
        self.add_field(DialogFieldLabel(""))
        self.adjust_limits_field = DialogFieldCheckbox(
            "Adjust min/max capacity if required",
            checked=True,
            color=Common.color("modal_dialog_textfield_label"),
            selected_color=Common.color("modal_dialog_textfield_selected"),
        )
        self.add_field(self.adjust_limits_field)
        self.caller = caller

    def accept_and_close(self):
        if self.desired_capacity_field.text == "":
            self.error_label.text = "Desired capacity cannot be blank."
            return

        b3s = Common.Session.service_provider("autoscaling")
        asg = b3s.describe_auto_scaling_groups(
            AutoScalingGroupNames=[self.asg_entry["name"]]
        )["AutoScalingGroups"][0]
        des = int(self.desired_capacity_field.text)

        if (
            des < asg["MinSize"] or des > asg["MaxSize"]
        ) and not self.adjust_limits_field.checked:
            self.error_label.text = (
                "Desired capacity is out of min-max range of {0}-{1}".format(
                    asg["MinSize"], asg["MaxSize"]
                )
            )
            return

        nmin = min(des, asg["MinSize"])
        nmax = max(des, asg["MaxSize"])

        b3s.update_auto_scaling_group(
            AutoScalingGroupName=self.asg_entry["name"],
            DesiredCapacity=des,
            MinSize=nmin,
            MaxSize=nmax,
        )
        super().accept_and_close()

    def close(self):
        if self.caller is not None:
            self.caller.refresh_data()
        super().close()


class ASGDescriber(Describer):
    prefix = "asg_browser"
    title = "Autoscaling Group"

    def __init__(
        self, parent, alignment, dimensions, entry, *args, entry_key="name", **kwargs
    ):
        self.resource_key = "autoscaling"
        self.describe_method = "describe_auto_scaling_groups"
        self.describe_kwarg_name = "AutoScalingGroupNames"
        self.describe_kwarg_is_list = True
        self.object_path = ".AutoScalingGroups[0]"
        super().__init__(
            parent,
            alignment,
            dimensions,
            *args,
            entry=entry,
            entry_key=entry_key,
            **kwargs
        )



import datetime
import os
import sys

from . import resources
from .aws import AWS
from .commander import Commander, Filterer
from .common import BaseChart, Common, DefaultAnchor, DefaultBorder, DefaultDimension
from .context import ContextList
from .info import InfoDisplay
from .log import LogLister, LogViewer
from .meta import CommanderOptionsLister
from .region import RegionList
from .ssh import SSHList
from .termui.alignment import Dimension, TopLeftAnchor
from .termui.control import Border


def awscheck():
    return bool(Common.Session.context) and bool(Common.Session.region)


def open_context_lister():
    ctxl = ContextList(
        Common.Session.ui.top_block,
        DefaultAnchor,
        DefaultDimension,
        border=DefaultBorder("context_list", "Contexts"),
        weight=0,
    )
    return [ctxl, ctxl.hotkey_display]


def open_region_lister():
    regl = RegionList(
        Common.Session.ui.top_block,
        DefaultAnchor,
        DefaultDimension,
        border=DefaultBorder("region_list", "Regions"),
        weight=0,
    )
    return [regl, regl.hotkey_display]


def open_ssh_lister():
    return SSHList.opener()


def open_logs_lister():
    return LogLister.opener()


def open_filterer():
    if Common.Session.filterer is None:
        return Filterer(
            Common.Session.ui.top_block,
            TopLeftAnchor(0, 8),
            Dimension("100%", "3"),
            Common.Session,
            color=Common.color("search_bar_color"),
            symbol_color=Common.color("search_bar_symbol_color"),
            autocomplete_color=Common.color("search_bar_autocomplete_color"),
            inactive_color=Common.color("search_bar_inactive_color"),
            weight=-200,
            border=Border(
                Common.border("search_bar"), Common.color("search_bar_border")
            ),
        )
    else:
        Common.Session.filterer.resume()


def open_commander():
    return Commander(
        Common.Session.ui.top_block,
        TopLeftAnchor(0, 8),
        Dimension("100%", "3"),
        Common.Session,
        color=Common.color("command_bar_color"),
        symbol_color=Common.color("command_bar_symbol_color"),
        autocomplete_color=Common.color("command_bar_autocomplete_color"),
        ok_color=Common.color("command_bar_ok_color"),
        error_color=Common.color("command_bar_error_color"),
        weight=-200,
        border=Border(Common.border("search_bar"), Common.color("search_bar_border")),
    )


def main(*args, **kwargs):
    # stderr hack
    old_stderr = None
    try:
        if os.fstat(0) == os.fstat(1):
            tg = open("error.log", "w", buffering=1)

            old_stderr = sys.stderr
            sys.stderr = tg

        Common.initialize()
        Common.Session.service_provider = AWS()
        Common.post_initialize()
        Common.Session.replace_frame(open_context_lister())

        Common.Session.info_display.commander_hook = open_commander
        Common.Session.info_display.filterer_hook = open_filterer

        Common.Session.commander_options["ctx"] = open_context_lister
        Common.Session.commander_options["context"] = open_context_lister
        Common.Session.commander_options["region"] = open_region_lister
        Common.Session.commander_options["ssh"] = open_ssh_lister
        Common.Session.commander_options["logs"] = open_logs_lister
        Common.Session.commander_options["?"] = CommanderOptionsLister.opener
        Common.Session.commander_options["help"] = CommanderOptionsLister.opener

        Common.main()
    finally:
        if old_stderr is not None:
            sys.stderr.close()
            sys.stderr = old_stderr



import datetime

from .base_control import ResourceLister
from .common import BaseChart, Common


class MetricLister(ResourceLister):
    prefix = "metric_list"
    title = "Metrics"

    def title_info(self):
        if self.dimension is not None:
            return "{0}={1}".format(self.dimensions[0], self.dimension[1])
        return None

    def __init__(
        self, *args, metric_namespace=None, metric_name=None, dimension=None, **kwargs
    ):
        self.resource_key = "cloudwatch"
        self.list_method = "list_metrics"
        self.item_path = ".Metrics"
        self.column_paths = {
            "namespace": ".Namespace",
            "name": ".MetricName",
        }
        self.imported_column_sizes = {
            "namespace": 16,
            "name": 64,
        }
        self.hidden_columns = {
            "dimension": self.add_dimension,
        }
        self.dimension = dimension
        self.list_kwargs = {}
        if metric_namespace is not None:
            self.list_kwargs["Namespace"] = metric_namespace
        if metric_name is not None:
            self.list_kwargs["MetricName"] = metric_name
        if dimension is not None:
            self.list_kwargs["Dimensions"] = [
                {"Name": dimension[0], "Value": dimension[1]}
            ]
        self.primary_key = "name"
        self.sort_column = "name"
        self.open_command = MetricViewer.opener
        self.open_selection_arg = "metric"
        super().__init__(*args, **kwargs)

    def add_dimension(self, entry):
        if self.dimension is not None:
            return "{0}={1}".format(self.dimension[0], self.dimension[1])
        return ""


class MetricViewer(BaseChart):
    prefix = "metric_view"
    title = "Metrics"

    def title_info(self):
        return "{0} {1}".format(self.metric_dimension["Value"], self.metric_name)

    def __init__(self, *args, metric=None, **kwargs):
        super().__init__(*args, **kwargs)
        self.metric_name = metric["name"]
        self.metric_namespace = metric["namespace"]
        d = metric["dimension"].split("=")
        self.metric_dimension = {
            "Name": d[0],
            "Value": "=".join(d[1:]),
        }
        self.load_data()

    def load_data(self):
        api_kwargs = {
            "MetricDataQueries": (
                [
                    {
                        "Id": "metric",
                        "MetricStat": {
                            "Metric": {
                                "Namespace": self.metric_namespace,
                                "MetricName": self.metric_name,
                                "Dimensions": [self.metric_dimension],
                            },
                            "Period": 300,
                            "Stat": "Average",
                        },
                    }
                ],
            ),
            "StartTime": (datetime.datetime.now() - datetime.timedelta(hours=24),),
            "EndTime": (datetime.datetime.now(),),
        }
        call = Common.generic_api_call(
            "cloudwatch",
            "get_metric_data",
            api_kwargs,
            "Retrieve metrics",
            "Cloudwatch",
            subcategory="Metrics",
            resource="{0}/{1}".format(self.metric_namespace, self.metric_name),
        )
        if call["Success"]:
            data = call["Response"]
            for idx in range(len(data["MetricDataResults"][0]["Timestamps"])):
                ts = data["MetricDataResults"][0]["Timestamps"][idx]
                val = data["MetricDataResults"][0]["Values"][idx]
                self.add_datapoint(ts, val)



from typing import Callable

from .termui.common import Commons
from .termui.control import Control
from .termui.dialog import DialogControl
from .termui.ui import ControlCodes


class InfoDisplay(Control):
    def __init__(
        self,
        parent,
        alignment,
        dimensions,
        *args,
        info=None,
        cols=2,
        highlight_color=None,
        generic_color=None,
        **kwargs
    ):
        super().__init__(parent, alignment, dimensions, *args, **kwargs)
        self.info = {}
        self.order = info[:] if info is not None else []
        self.special_colors = {}
        for k in info:
            self.info[k] = None
        self.cols = cols
        self.highlight_color = highlight_color
        self.generic_color = generic_color
        self.commander_hook: Callable[[], None] = None
        self.filterer_hook: Callable[[], None] = None

    def input(self, key):
        if key == ":" and self.commander_hook is not None:
            # pylint: disable=not-callable # Member is callable or None, and check asserts that it is not None
            self.commander_hook()
            return True
        if key == "/" and self.filterer_hook is not None:
            # pylint: disable=not-callable # Member is callable or None, and check asserts that it is not None
            self.filterer_hook()
            return True
        return False

    def __getitem__(self, k):
        return self.info[k] if k in self.info else None

    def __setitem__(self, k, v):
        if k not in self.order:
            self.order.append(k)
        self.info[k] = v

    def paint(self):
        super().paint()
        (x0, x1), (y0, y1) = self.inner
        w = x1 - x0 + 1
        colw = int(w / self.cols)
        x = x0
        y = y0
        longest = []
        v = 0

        for name in self.order:
            if name not in self.info:
                continue
            display = name + ": "
            if len(display) > v:
                v = len(display)
            y += 1
            if y > y1:
                y = y0
                longest.append(v)
                v = 0
        if v > 0:
            longest.append(v)
        y = y0
        col = 0
        for name in self.order:
            if name not in self.info:
                continue
            value = self.info[name]
            if value is None:
                value = ""
            display = name + ": "
            if len(display) < longest[col]:
                display += " " * (longest[col] - len(display))
            Commons.UIInstance.print(
                display, xy=(x, y), color=self.highlight_color, bold=True
            )
            t = value
            if len(t) > colw - longest[col]:
                t = t[: colw - longest[col]]
            Commons.UIInstance.print(
                t,
                xy=(x + longest[col], y),
                color=self.generic_color
                if name not in self.special_colors
                else self.special_colors[name],
            )
            y += 1
            if y > y1:
                y = y0
                col += 1
                x += colw


class NeutralDialog(DialogControl):
    def input(self, key):
        return False


class HotkeyDisplay(Control):
    translations = {
        "KEY_DOWN": "↓",
        "KEY_RIGHT": "→",
        "KEY_UP": "↑",
        "KEY_LEFT": "←",
        "KEY_ENTER": "↲",
        "\n": "↲",
        "\t": "↹",
        "KEY_BACKSPACE": "⇦",
        "KEY_DELETE": "del",
        "KEY_END": "end",
        "KEY_HOME": "home",
        "KEY_ESCAPE": "esc",
        "KEY_PGUP": "pgup",
        "KEY_PGDOWN": "pgdn",
        "KEY_INSERT": "ins",
        ControlCodes.A: "ctrl-a",
        ControlCodes.B: "ctrl-b",
        ControlCodes.C: "ctrl-c",
        ControlCodes.D: "ctrl-d",
        ControlCodes.E: "ctrl-e",
        ControlCodes.F: "ctrl-f",
        ControlCodes.G: "ctrl-g",
        ControlCodes.K: "ctrl-k",
        ControlCodes.L: "ctrl-l",
        ControlCodes.M: "ctrl-m",
        ControlCodes.N: "ctrl-n",
        ControlCodes.O: "ctrl-o",
        ControlCodes.P: "ctrl-p",
        ControlCodes.Q: "ctrl-q",
        ControlCodes.R: "ctrl-r",
        ControlCodes.S: "ctrl-s",
        ControlCodes.T: "ctrl-t",
        ControlCodes.U: "ctrl-u",
        ControlCodes.V: "ctrl-v",
        ControlCodes.W: "ctrl-w",
        ControlCodes.X: "ctrl-x",
        ControlCodes.Y: "ctrl-y",
        ControlCodes.Z: "ctrl-z",
    }

    def __init__(
        self,
        parent,
        alignment,
        dimensions,
        holder,
        *args,
        session=None,
        cols=2,
        highlight_color=None,
        generic_color=None,
        **kwargs
    ):
        super().__init__(parent, alignment, dimensions, *args, **kwargs)
        self.holder = holder
        self.cols = cols
        self.highlight_color = highlight_color
        self.generic_color = generic_color
        self.session = session

    def paint(self):
        super().paint()
        c = self.corners()
        x0 = c[0][0] + (0 if self.border is None else 1)
        x1 = c[0][1] - (0 if self.border is None else 1)
        y0 = c[1][0] + (0 if self.border is None else 1)
        y1 = c[1][1] - (0 if self.border is None else 1)
        w = x1 - x0 + 1
        colw = int(w / self.cols)
        x = x0
        y = y0
        longest = []
        v = 0
        tooltips = {**self.holder.tooltips, **self.session.global_hotkey_tooltips}
        for hotkey, tooltip in tooltips.items():
            display = (
                "<"
                + (
                    HotkeyDisplay.translations[hotkey]
                    if hotkey in HotkeyDisplay.translations
                    else hotkey
                )
                + "> "
            )
            if len(display) > v:
                v = len(display)
            y += 1
            if y > y1:
                y = y0
                longest.append(v)
                v = 0
        if v > 0:
            longest.append(v)
        y = y0
        col = 0
        for hotkey, tooltip in tooltips.items():
            display = (
                "<"
                + (
                    HotkeyDisplay.translations[hotkey]
                    if hotkey in HotkeyDisplay.translations
                    else hotkey
                )
                + "> "
            )
            if len(display) < longest[col]:
                display += " " * (longest[col] - len(display))
            Commons.UIInstance.print(
                display, xy=(x, y), color=self.highlight_color, bold=True
            )
            t = tooltip if not callable(tooltip) else tooltip()
            if len(t) > colw - longest[col]:
                t = t[: colw - longest[col]]
            Commons.UIInstance.print(
                t, xy=(x + longest[col], y), color=self.generic_color
            )
            y += 1
            if y > y1:
                y = y0
                col += 1
                x += colw



from .base_control import Describer, ResourceLister


class EBSResourceLister(ResourceLister):
    prefix = "ebs_list"
    title = "EBS Volumes"
    command_palette = ["ebs"]

    def __init__(self, *args, **kwargs):
        self.resource_key = "ec2"
        self.list_method = "describe_volumes"
        self.item_path = ".Volumes"
        self.column_paths = {
            "id": ".VolumeId",
            "name": self.tag_finder_generator("Name"),
            "size": self.determine_size,
            "state": ".State",
            "type": ".VolumeType",
            "attached to": self.determine_attachment,
        }
        self.imported_column_sizes = {
            "id": 20,
            "name": 32,
            "size": 10,
            "state": 10,
            "type": 10,
            "attached to": 16,
        }
        self.imported_column_order = [
            "id",
            "name",
            "size",
            "state",
            "type",
            "attached to",
        ]
        self.list_kwargs = {}
        self.primary_key = "id"
        self.sort_column = "id"
        self.describe_command = EBSDescriber.opener
        super().__init__(*args, **kwargs)

    def determine_size(self, entry):
        return "{0} GiB".format(entry["Size"])

    def determine_attachment(self, entry):
        for at in entry["Attachments"]:
            if at["State"] in ("attaching", "attached", "busy"):
                return at["InstanceId"]
        return ""


class EBSDescriber(Describer):
    prefix = "ebs_browser"
    title = "EBS Volume"

    def __init__(
        self, parent, alignment, dimensions, entry, *args, entry_key="id", **kwargs
    ):
        self.resource_key = "ec2"
        self.describe_method = "describe_volumes"
        self.describe_kwarg_name = "VolumeIds"
        self.describe_kwarg_is_list = True
        self.object_path = ".Volumes[0]"
        super().__init__(
            parent,
            alignment,
            dimensions,
            *args,
            entry=entry,
            entry_key=entry_key,
            **kwargs
        )



from .base_control import Describer, ResourceLister


class VPCResourceLister(ResourceLister):
    prefix = "vpc_list"
    title = "VPCs"
    command_palette = ["vpc"]

    def __init__(self, *args, **kwargs):
        from .resource_subnet import SubnetResourceLister

        self.resource_key = "ec2"
        self.list_method = "describe_vpcs"
        self.item_path = ".Vpcs"
        self.column_paths = {
            "id": ".VpcId",
            "name": self.tag_finder_generator("Name"),
            "default": self.determine_default,
            "cidr": ".CidrBlock",
            "state": ".State",
        }
        self.imported_column_sizes = {
            "id": 30,
            "name": 30,
            "default": 3,
            "cidr": 18,
            "state": 9,
        }
        self.describe_command = VPCDescriber.opener
        self.additional_commands = {
            "t": {
                "command": SubnetResourceLister.opener,
                "selection_arg": "vpc",
                "tooltip": "View Subnets",
            }
        }
        self.open_command = "t"

        self.imported_column_order = ["id", "name", "default", "cidr", "state"]
        self.sort_column = "id"
        self.primary_key = "id"
        super().__init__(*args, **kwargs)

    def determine_default(self, vpc, *args):
        if vpc["IsDefault"]:
            return "✓"
        return ""


class VPCDescriber(Describer):
    prefix = "vpc_browser"
    title = "VPC"

    def __init__(
        self, parent, alignment, dimensions, entry, *args, entry_key="id", **kwargs
    ):
        self.resource_key = "ec2"
        self.describe_method = "describe_vpcs"
        self.describe_kwarg_name = "VpcIds"
        self.describe_kwarg_is_list = True
        self.object_path = ".Vpcs[0]"
        super().__init__(
            parent,
            alignment,
            dimensions,
            *args,
            entry=entry,
            entry_key=entry_key,
            **kwargs
        )



import datetime

from .color import ColorDarkGreen, ColorGold
from .common import Commons
from .control import Control


class BarGraph(Control):
    DPB_ZeroOrLessToMax = 0
    DPB_ZeroToMax = 1
    DPB_MinToMax = 2

    def __init__(
        self,
        parent,
        alignment,
        dimensions,
        *args,
        color=ColorGold,
        highlight_color=ColorDarkGreen,
        **kwargs
    ):
        super().__init__(parent, alignment, dimensions, *args, **kwargs)
        self.subdivisions = {
            0: " ",
            1: "▁",
            2: "▂",
            3: "▃",
            4: "▄",
            5: "▅",
            6: "▆",
            7: "▇",
            8: "█",
            -1: "▔",
            -2: "▔",
            -3: "▔",
            -4: "▀",
            -5: "▀",
            -6: "▀",
            -7: "▀",
            -8: "█",
            "?": "░",
        }

        self._data_points_behaviour = BarGraph.DPB_ZeroOrLessToMax

        self.datapoints = []
        self.max_point = 0
        self.min_point = 0
        now = datetime.datetime.now()
        self.min_time = now
        self.max_time = now
        self.color = color
        self.highlight_color = highlight_color
        self.hotkeys = {
            "KEY_END": self.end,
            "KEY_HOME": self.home,
            "KEY_LEFT": self.scroll_left,
            "KEY_RIGHT": self.scroll_right,
        }
        self.tooltips = {}

        self.left = 0
        self.highlight = 0

    def input(self, key):
        inkey = str(key)
        if key.is_sequence:
            inkey = key.name
        else:
            inkey = inkey.lower()
        if inkey in self.hotkeys.keys():
            self.hotkeys[inkey](self)
            Commons.UIInstance.dirty = True
            return True
        return False

    def add_hotkey(self, hotkey, action, tooltip=None):
        self.hotkeys[hotkey] = action
        if tooltip is not None:
            self.tooltips[hotkey] = tooltip
        Commons.UIInstance.dirty = True

    def scroll_left(self, *args):
        if self.highlight > 0:
            self.highlight -= 1
            if self.left > self.highlight:
                self.left = self.highlight
            Commons.UIInstance.dirty = True

    def home(self, *args):
        self.left = 0
        self.highlight = 0
        Commons.UIInstance.dirty = True

    def end(self, *args):
        self.highlight = len(self.datapoints) - 1
        self.left = max(len(self.datapoints) - self.colspace, 0)
        Commons.UIInstance.dirty = True

    def scroll_right(self, *args):
        if self.highlight < len(self.datapoints) - 1:
            self.highlight += 1
            if self.highlight >= self.left + self.colspace:
                self.left += 1
            Commons.UIInstance.dirty = True

    @property
    def data_points_behaviour(self):
        return self._data_points_behaviour

    @data_points_behaviour.setter
    def data_points_behaviour(self, value):
        self._data_points_behaviour = value
        if value == BarGraph.DPB_ZeroOrLessToMax:
            self.min_point = 0
            for dp in self.datapoints:
                if dp[1] < self.min_point:
                    self.min_point = dp[1]
        elif value == BarGraph.DPB_ZeroToMax:
            self.min_point = 0
        elif value == BarGraph.DPB_MinToMax:
            self.min_point = self.max_point
            for dp in self.datapoints:
                if dp[1] < self.min_point:
                    self.min_point = dp[1]
        else:
            raise ValueError

    def add_datapoint(self, timestamp, datapoint):
        try:
            if len(self.datapoints) == 0:
                self.max_point = datapoint
                if self._data_points_behaviour == BarGraph.DPB_ZeroOrLessToMax:
                    self.min_point = min(0, datapoint)
                elif self._data_points_behaviour == BarGraph.DPB_ZeroToMax:
                    self.min_point = 0
                elif self._data_points_behaviour == BarGraph.DPB_MinToMax:
                    self.min_point = datapoint
            else:
                self.max_point = max(self.max_point, datapoint)
                if self._data_points_behaviour == BarGraph.DPB_ZeroOrLessToMax:
                    self.min_point = min(self.min_point, min(0, datapoint))
                elif self._data_points_behaviour == BarGraph.DPB_ZeroToMax:
                    self.min_point = 0
                elif self._data_points_behaviour == BarGraph.DPB_MinToMax:
                    self.min_point = min(self.min_point, datapoint)
            for i in range(len(self.datapoints)):
                ts = self.datapoints[i][0]
                if ts > timestamp:
                    self.datapoints.insert(i, (timestamp, datapoint))
                    return
            self.datapoints.append((timestamp, datapoint))
        finally:
            self.left = max(len(self.datapoints) - self.colspace, 0)
            self.highlight = len(self.datapoints) - 1
            Commons.UIInstance.dirty = True

    @Control.border.setter  # type: ignore[attr-defined]
    def border(self, value):
        self._border = value
        # Do not execute on border initialization
        if hasattr(self, "datapoints"):
            self.left = max(len(self.datapoints) - self.colspace, 0)
            Commons.UIInstance.dirty = True

    @property
    def colspace(self):
        c = self.corners()
        wspace = c[0][1] - c[0][0] + 1 - (0 if self.border is None else 2)
        max_label = "{0:.2f}".format(float(self.max_point))
        min_label = "{0:.2f}".format(float(self.min_point))
        x_label_space = max(len(max_label), len(min_label)) + 1
        return wspace - x_label_space

    def paint(self):
        super().paint()
        c = self.corners()
        wspace = c[0][1] - c[0][0] + 1 - (0 if self.border is None else 2)
        hspace = c[1][1] - c[1][0] + 1 - (0 if self.border is None else 2)
        x0 = c[0][0] + (0 if self.border is None else 1)
        y0 = c[1][0] + (0 if self.border is None else 1)

        max_label = "{0:.2f}".format(float(self.max_point))
        min_label = "{0:.2f}".format(float(self.min_point))

        x_label_space = max(len(max_label), len(min_label)) + 1
        colspace = wspace - x_label_space
        if colspace == 0:
            return
        first_col = self.left
        last_col = min(self.left + colspace - 1, len(self.datapoints) - 1)
        col_width = float(colspace) / float((last_col - first_col + 1))
        roll = col_width - int(col_width)
        col_width = int(col_width)
        rolling = 0

        rowspace = hspace - 2
        if rowspace == 0:
            return
        point_step = (self.max_point - self.min_point) / rowspace

        xb0 = x0 + x_label_space

        Commons.UIInstance.print(max_label, (x0, y0), color=self.color)
        Commons.UIInstance.print(min_label, (x0, y0 + rowspace - 1), color=self.color)
        Commons.UIInstance.print(
            self.datapoints[first_col][0].isoformat(),
            (xb0, y0 + rowspace),
            color=self.color,
        )
        last_ts = self.datapoints[last_col][0].isoformat()
        Commons.UIInstance.print(
            last_ts,
            (xb0 + colspace - len(last_ts) - 1, y0 + rowspace),
            color=self.color,
        )
        Commons.UIInstance.print(
            "Value at {0}: {1:.2f}".format(
                self.datapoints[self.highlight][0].isoformat(),
                self.datapoints[self.highlight][1],
            ),
            (xb0, y0 + rowspace + 1),
            color=self.highlight_color,
        )

        v = self.max_point
        for y in range(y0, y0 + rowspace):
            row_bounds = (v - point_step, v)
            v -= point_step
            row = ["", "", ""]
            ri = 0
            rolling = 0
            for col in range(first_col, last_col + 1):
                datapoint = self.datapoints[col][1]
                colw = col_width
                if col == self.highlight:
                    ri = 1
                if rolling > 1:
                    colw += 1
                    rolling -= 1
                if datapoint < row_bounds[0]:
                    if row_bounds[0] > 0:
                        row[ri] += self.subdivisions[0] * colw
                    else:
                        row[ri] += self.subdivisions[-8] * colw
                elif datapoint > row_bounds[1]:
                    if row_bounds[1] > 0:
                        row[ri] += self.subdivisions[8] * colw
                    else:
                        row[ri] += self.subdivisions[0] * colw
                else:
                    if row_bounds[0] < 0 and row_bounds[1] > 0:
                        row[ri] += self.subdivisions[8] * colw
                    elif datapoint > 0:
                        if point_step != 0:
                            t = int((float(datapoint - row_bounds[0]) / point_step) * 8)
                        else:
                            t = "?"
                        row[ri] += self.subdivisions[t] * colw
                    else:
                        if point_step != 0:
                            t = int((float(datapoint - row_bounds[1]) / point_step) * 8)
                        else:
                            t = "?"
                        row[ri] += self.subdivisions[t] * colw
                rolling += roll
                if col == self.highlight:
                    ri = 2
            Commons.UIInstance.print(row[0], (xb0, y), color=self.color)
            if len(row[1]) > 0:
                Commons.UIInstance.print(
                    row[1], (xb0 + len(row[0]), y), color=self.highlight_color
                )
                if len(row[2]) > 0:
                    Commons.UIInstance.print(
                        row[2], (xb0 + len(row[0]) + len(row[1]), y), color=self.color
                    )



from pathlib import Path

from .base_control import Describer, GenericDescriber, ResourceLister
from .common import Common
from .termui.list_control import ListEntry
from .termui.ui import ControlCodes


class S3ResourceLister(ResourceLister):
    prefix = "s3_list"
    title = "S3 Buckets"
    command_palette = ["s3"]

    def __init__(self, *args, **kwargs):
        self.resource_key = "s3"
        self.list_method = "list_buckets"
        self.item_path = ".Buckets"
        self.column_paths = {
            "name": ".Name",
            "location": self.determine_location,
        }
        self.imported_column_sizes = {
            "name": 60,
            "location": 20,
        }
        self.describe_command = S3Describer.opener
        self.open_command = S3ObjectLister.opener
        self.open_selection_arg = "bucket"
        self.primary_key = "name"

        self.imported_column_order = ["name", "location"]
        self.sort_column = "name"
        super().__init__(*args, **kwargs)

    def determine_location(self, entry):
        loc_resps = Common.generic_api_call(
            "s3",
            "get_bucket_location",
            {"Bucket": entry["name"]},
            "Get Bucket Location",
            "S3",
            subcategory="Bucket",
            resource=entry["name"],
        )
        if loc_resps["Success"]:
            loc_resp = loc_resps["Response"]
            return loc_resp["LocationConstraint"]
        return "<n/a>"


class S3Describer(Describer):
    prefix = "s3_browser"
    title = "S3 Bucket"

    def __init__(
        self, parent, alignment, dimensions, entry, *args, entry_key="name", **kwargs
    ):
        self.resource_key = "rds"
        self.describe_method = "get_bucket_policy"
        self.describe_kwarg_name = "Bucket"
        self.object_path = "."
        super().__init__(
            parent,
            alignment,
            dimensions,
            *args,
            entry=entry,
            entry_key=entry_key,
            **kwargs
        )

    def title_info(self):
        return self.bucket


class CancelDownload(Exception):
    pass


class S3ObjectLister(ResourceLister):
    prefix = "s3_object_list"
    title = "S3 Browser"

    def title_info(self):
        if self.path is None:
            return self.bucket
        else:
            return "{0}/{1}".format(self.bucket, self.path)

    def __init__(self, *args, bucket, path=None, **kwargs):
        self.icons = {
            "folder": "🗀",
            "default": "🗈",
            "ext.txt": "🖹",
            "ext.png": "🖻",
            "ext.jpg": "🖻",
            "ext.gif": "🖻",
            "ext.jpeg": "🖻",
            "ext.svg": "🖻",
            "ext.doc": "🖺",
            "ext.docx": "🖺",
            "ext.xls": "🖺",
            "ext.xlsx": "🖺",
            "ext.ppt": "🖺",
            "ext.pptx": "🖺",
        }

        self.resource_key = "s3"
        self.list_method = "list_objects"
        if isinstance(bucket, ListEntry):
            self.bucket = bucket["name"]
        else:
            self.bucket = bucket
        self.path = path
        self.list_kwargs = {"Bucket": self.bucket, "MaxKeys": 100, "Delimiter": "/"}
        self.next_marker = "NextMarker"
        self.next_marker_arg = "Marker"
        if self.path is not None:
            self.list_kwargs["Prefix"] = self.path
        self.item_path = ".Contents + .CommonPrefixes"
        self.column_paths = {
            "icon": self.determine_icon,
            "name": self.determine_name,
            "size": self.determine_size,
        }
        self.hidden_columns = {
            "is_dir": self.determine_is_dir,
            "size_in_bytes": ".Size",
            "bucket_prefixed_path": self.determine_bucket_prefixed_path,
        }
        self.imported_column_sizes = {
            "icon": 1,
            "name": 150,
            "size": 20,
        }
        self.imported_column_order = ["icon", "name", "size"]
        self.sort_column = "name"
        self.additional_commands = {
            "d": {
                "command": self.download_selection,
                "selection_arg": "file",
                "tooltip": "Download",
            },
            "v": {
                "command": self.view_selection,
                "selection_arg": "file",
                "tooltip": "View Contents",
            },
        }
        self.open_command = S3ObjectLister.opener
        self.open_selection_arg = "path"
        super().__init__(*args, **kwargs)
        self.primary_key = "bucket_prefixed_path"
        self.add_hotkey(ControlCodes.D, self.empty, "Cancel download")

    def determine_bucket_prefixed_path(self, entry, *args):
        return "{0}/{1}".format(
            self.bucket, entry["Key"] if "Key" in entry else entry["Prefix"]
        )

    def determine_name(self, entry, *args):
        if "Prefix" in entry:
            return entry["Prefix"].strip("/").split("/")[-1]
        return entry["Key"].split("/")[-1]

    def determine_is_dir(self, entry, *args):
        return "/" if "Prefix" in entry else ""

    def determine_size(self, entry, *args):
        if "Prefix" in entry:
            return ""
        b_prefix = ["", "Ki", "Mi", "Gi", "Ti", "Ei"]
        b_idx = 0
        s = float(entry["Size"])
        while s >= 1024:
            b_idx += 1
            s /= 1024
            if b_idx == len(b_prefix) - 1:
                break
        return "{0:.2f} {1}B".format(s, b_prefix[b_idx])

    def determine_icon(self, entry, *args):
        if "Prefix" in entry:
            return self.icons["folder"]
        ext = entry["Key"].split("/")[-1].split(".")[-1]
        ext_key = "ext.{0}".format(ext)
        if ext_key in self.icons:
            return self.icons[ext_key]
        return self.icons["default"]

    def open(self, *args):
        if (
            self.open_command is not None
            and self.selection is not None
            and self.selection["is_dir"] == "/"
        ):
            subpath = self.get_selection_path()
            self.command(
                self.open_command,
                {
                    self.open_selection_arg: subpath,
                    "bucket": self.bucket,
                    "pushed": True,
                },
            )

    def get_selection_path(self):
        if self.selection is not None:
            if self.path is None:
                return "{0}{1}".format(self.selection["name"], self.selection["is_dir"])
            else:
                return "{0}{1}{2}".format(
                    self.path, self.selection["name"], self.selection["is_dir"]
                )
        return None

    def get_cache_name(self):
        return "{0}/{1}".format(self.bucket, self.get_selection_path()).replace(
            "/", "--"
        )

    def cache_fetch(self, obj, *args, **kwargs):
        if self.selection is not None:
            sp = self.get_selection_path()
            obj_size = float(self.selection["size_in_bytes"])
            downloaded = 0

            def fn(chunk):
                nonlocal downloaded
                downloaded += chunk
                perc = float(downloaded) / obj_size
                key = Common.Session.ui.check_one_key()
                if key == ControlCodes.C:
                    raise KeyboardInterrupt
                elif key == ControlCodes.D:
                    raise CancelDownload()
                Common.Session.ui.progress_bar_paint(perc)

            Common.generic_api_call(
                "s3",
                "download_file",
                {"Bucket": self.bucket, "Key": sp, "Filename": obj, "Callback": fn},
                "Download File",
                "S3",
                subcategory="Object",
                resource="{0}/{1}".format(self.bucket, sp),
            )

            # Common.Session.service_provider("s3").download_file(
            #    Bucket=self.bucket, Key=sp, Filename=obj, Callback=fn
            # )
            try:
                with open(obj, "r") as f:
                    return f.read()
            except UnicodeDecodeError:
                with open(obj, "rb") as f:
                    return f.read()

    def download_selection(self, *args, **kwargs):
        if self.selection is not None:
            if self.selection["is_dir"] == "/":
                Common.Session.set_message(
                    "Cannot download directory as file.", Common.color("message_info")
                )
                return
            path = Path.home() / "Downloads" / "awsc"
            path.mkdir(parents=True, exist_ok=True)
            try:
                data = Common.Session.ui.filecache(
                    self.get_cache_name(), self.cache_fetch
                )
            except CancelDownload:
                Common.Session.set_message(
                    "Download cancelled", Common.color("message_info")
                )
                return None
            fpath = path / self.selection["name"]
            mode = "w"
            if isinstance(data, bytes):
                mode = "wb"
            with fpath.open(mode) as f:
                f.write(data)
            Common.Session.set_message(
                "Downloaded file to {0}".format(str(fpath.resolve())),
                Common.color("message_success"),
            )
        return None

    def view_selection(self, *args, **kwargs):
        if self.selection is not None:
            if self.selection["is_dir"] == "/":
                Common.Session.set_message(
                    "Cannot display directory as text file.",
                    Common.color("message_info"),
                )
                return None
            try:
                data = Common.Session.ui.filecache(
                    self.get_cache_name(), self.cache_fetch
                )
            except CancelDownload:
                Common.Session.set_message(
                    "Download cancelled", Common.color("message_info")
                )
                return None
            if isinstance(data, bytes):
                data = "<binary>"
            return GenericDescriber.opener(
                **{
                    "describing": "File: {0}".format(self.get_selection_path()),
                    "content": data,
                    "pushed": True,
                }
            )
        return None



from .base_control import (
    DeleteResourceDialog,
    Describer,
    ListResourceDocumentCreator,
    ResourceLister,
    SingleRelationLister,
    SingleSelectorDialog,
)
from .common import Common
from .termui.alignment import CenterAnchor, Dimension
from .termui.dialog import DialogFieldCheckbox
from .termui.ui import ControlCodes


class SQSLister(ResourceLister):
    prefix = "sqs_list"
    title = "Queues"
    command_palette = ["sqs", "queue", "queues"]

    def delete_queue(self, _):
        if self.selection is None:
            return
        DeleteResourceDialog(
            self.parent,
            CenterAnchor(0, 0),
            Dimension("80%|40", "10"),
            caller=self,
            resource_type="queue",
            resource_identifier=self.selection["url"],
            callback=self.do_delete,
            action_name="Delete",
        )

    def do_delete(self, **kwargs):
        if self.selection is None:
            return
        Common.generic_api_call(
            "sqs",
            "delete_queue",
            {"QueueUrl": self.selection["url"]},
            "Delete Queue",
            "SQS",
            subcategory="Queue",
            success_template="Deleting queue {0}",
            resource=self.selection["url"],
        )
        self.refresh_data()

    def purge_queue(self, _):
        if self.selection is None:
            return
        DeleteResourceDialog(
            self.parent,
            CenterAnchor(0, 0),
            Dimension("80%|40", "10"),
            caller=self,
            resource_type="queue",
            resource_identifier=self.selection["url"],
            callback=self.do_purge,
            action_name="Purge",
        )

    def do_purge(self, force, terminate_resources_field, **kwargs):
        if self.selection is None:
            return
        Common.generic_api_call(
            "sqs",
            "purge_queue",
            {"QueueUrl": self.selection["url"]},
            "Purge Queue",
            "SQS",
            subcategory="Queue",
            success_template="Purging queue {0}",
            resource=self.selection["url"],
        )
        self.refresh_data()

    def send_message(self, _):
        if self.selection is None:
            return
        doc = {
            "MessageBody": "",
            "DelaySeconds": 0,
            "MessageAttributes": {},
            "MessageSystemAttributes": {},
        }
        if self.selection["arn"].endswith(".fifo"):
            doc["MessageDeduplicationId"]: ""
            doc["MessageGroupId"]: ""
        creator = ListResourceDocumentCreator(
            "sqs",
            "send_message",
            None,
            as_json=["MessageAttributes", "MessageSystemAttributes"],
            initial_document=doc,
            static_fields={"QueueUrl": self.selection["url"]},
            message="Message successfully sent",
        )
        creator.edit()

    def create_queue(self, _):
        doc = {
            "QueueName": "",
            "Attributes": {
                "DelaySeconds": 0,
                "MaximumMessageSize": 262144,
                "MessageRetentionPeriod": 345600,
                "Policy": "",
                "ReceiveMessageWaitTimeSeconds": 0,
                "RedrivePolicy": "",
                "VisibilityTimeout": 30,
                "KmsMasterKeyId": "",
                "KmsDataKeyReusePeriodSeconds": 300,
                "SqsManagedSseEnabled": False,
                "FifoQueue": False,
                "ContentBasedDeduplication": False,
                "DeduplicationScope": "queue",
                "FifoThroughputLimit": "perQueue",
            },
            "tags": {},
        }
        creator = ListResourceDocumentCreator(
            "sqs",
            "create_queue",
            None,
            as_json=["Attributes", "tags"],
            initial_document=doc,
        )
        creator.edit()

    def __init__(self, *args, **kwargs):
        self.resource_key = "sqs"
        self.list_method = "list_queues"
        self.item_path = ".QueueUrls"
        self.column_paths = {"url": ".", "available": self.determine_available_messages}
        self.imported_column_sizes = {
            "url": 80,
            "available": 9,
        }
        self.hidden_columns = {"arn": self.determine_queue_arn, "name": "."}
        self.describe_command = SQSDescriber.opener

        self.imported_column_order = ["url", "available"]
        self.sort_column = "url"
        self.primary_key = "url"
        super().__init__(*args, **kwargs)
        self.add_hotkey(ControlCodes.D, self.delete_queue, "Delete Queue")
        self.add_hotkey(ControlCodes.P, self.purge_queue, "Purge Queue")
        self.add_hotkey(ControlCodes.S, self.send_message, "Send Message")
        self.add_hotkey(ControlCodes.N, self.create_queue, "Create Queue")
        self._attrib_cache = {}

    def refresh_data(self, *args, **kwargs):
        self._attrib_cache = {}
        return super().refresh_data(*args, **kwargs)

    def get_attrib_cache(self, url, attrib):
        if url not in self._attrib_cache:
            resps = Common.generic_api_call(
                "sqs",
                "get_queue_attributes",
                {"QueueUrl": url, "AttributeNames": ["All"]},
                "Get Queue Attributes",
                "SQS",
                subcategory="Queue",
                resource=url,
            )
            if resps["Success"]:
                attribs = resps["Response"]
                self._attrib_cache[url] = attribs
        try:
            return self._attrib_cache[url]["Attributes"][attrib]
        except KeyError:
            return "<n/a>"

    def determine_queue_arn(self, v):
        return self.get_attrib_cache(v, "QueueArn")

    def determine_available_messages(self, v):
        return self.get_attrib_cache(v, "ApproximateNumberOfMessages")


class SQSDescriber(Describer):
    prefix = "sqs_browser"
    title = "Queue"

    def __init__(
        self, parent, alignment, dimensions, entry, *args, entry_key="url", **kwargs
    ):
        self.resource_key = "sqs"
        self.describe_method = "get_queue_attributes"
        self.describe_kwarg_name = "QueueUrl"
        self.describe_kwargs = {"AttributeNames": ["All"]}
        self.object_path = "."
        super().__init__(
            parent,
            alignment,
            dimensions,
            *args,
            entry=entry,
            entry_key=entry_key,
            **kwargs
        )



from .base_control import Describer, ResourceLister
from .common import Common


class SubnetResourceLister(ResourceLister):
    prefix = "subnet_list"
    title = "Subnets"
    command_palette = ["subnet"]

    def title_info(self):
        if self.vpc is not None:
            return self.vpc["id"]
        elif self.db_subnet_group is not None:
            return self.db_subnet_group["name"]
        return None

    def __init__(self, *args, **kwargs):
        from .resource_routing import RouteTableResourceLister

        self.resource_key = "ec2"
        self.list_method = "describe_subnets"
        self.item_path = ".Subnets"
        if "vpc" in kwargs:
            self.vpc = kwargs["vpc"]
            self.list_kwargs = {
                "Filters": [
                    {
                        "Name": "vpc-id",
                        "Values": [
                            self.vpc["id"],
                        ],
                    }
                ]
            }
        else:
            self.vpc = None
        if "db_subnet_group" in kwargs:
            self.db_subnet_group = kwargs["db_subnet_group"]
            self.list_kwargs = self.get_db_subnet_ids
        else:
            self.db_subnet_group = None
        self.column_paths = {
            "id": ".SubnetId",
            "name": self.tag_finder_generator("Name"),
            "vpc": ".VpcId",
            "cidr": ".CidrBlock",
            "AZ": ".AvailabilityZone",
            "public": self.determine_public,
        }
        self.hidden_columns = {
            "arn": ".SubnetArn",
        }
        self.imported_column_sizes = {
            "id": 30,
            "name": 30,
            "vpc": 30,
            "cidr": 18,
            "AZ": 20,
            "public": 3,
        }
        self.describe_command = SubnetDescriber.opener
        self.additional_commands = {
            "t": {
                "command": RouteTableResourceLister.opener,
                "selection_arg": "subnet",
                "tooltip": "View Route Table",
            }
        }
        # self.open_command = RouteTableResourceLister.opener
        # self.open_selection_arg = 'subnet'

        self.imported_column_order = ["id", "name", "vpc", "cidr", "AZ", "public"]
        self.sort_column = "id"
        self.primary_key = "id"
        super().__init__(*args, **kwargs)

    def determine_public(self, subnet, *args):
        if subnet["MapPublicIpOnLaunch"]:
            return "✓"
        return ""

    def get_db_subnet_ids(self, *args):
        dsg_resp = Common.generic_api_call(
            "rds",
            "describe_db_subnet_groups",
            {"DBSubnetGroupName": self.db_subnet_group.name},
            "Describe DB Subnet Group",
            "RDS",
            subcategory="Subnet Group",
            resource=self.db_subnet_group.name,
        )
        if not dsg_resp["Success"]:
            return {"SubnetIds": []}
        dsg = dsg_resp["Response"]
        return {
            "SubnetIds": [
                s["SubnetIdentifier"] for s in dsg["DBSubnetGroups"][0]["Subnets"]
            ]
        }


class SubnetDescriber(Describer):
    prefix = "subnet_browser"
    title = "Subnet"

    def __init__(
        self, parent, alignment, dimensions, entry, *args, entry_key="id", **kwargs
    ):
        self.resource_key = "ec2"
        self.describe_method = "describe_subnets"
        self.describe_kwarg_name = "SubnetIds"
        self.describe_kwarg_is_list = True
        self.object_path = ".Subnets[0]"
        super().__init__(
            parent,
            alignment,
            dimensions,
            *args,
            entry=entry,
            entry_key=entry_key,
            **kwargs
        )



class Character:
    def __init__(self):
        self.value = " "
        self.color = Character.nullcolor
        self.bold = False
        self.out = " "
        self.dirty = False

    @staticmethod
    def nullcolor(x, **kwargs):
        return x

    def clear(self):
        self.out = "\033[38;5;220m \033[0m"
        self.dirty = False

    def output(self):
        if self.dirty:
            self.dirty = False
            o = self.value
            if self.color is not None:
                o = self.color(o, bold=self.bold)
            self.out = o
        return self.out


class Screen:
    def __init__(self, ui):
        self.ui = ui
        self.buf = []
        self.clear()

    class Row:
        def __init__(self, screen):
            self.screen = screen
            self.buf = []
            self.clear()

        def clear(self):
            if len(self.buf) == self.screen.ui.w:
                for c in self.buf:
                    c.clear()
            else:
                self.buf = [Character() for i in range(self.screen.ui.w)]

        def __getitem__(self, i):
            return self.buf[i]

        def output(self):
            return "".join(x.output() for x in self.buf)

    def clear(self):
        if len(self.buf) == self.ui.h:
            for c in self.buf:
                c.clear()
        else:
            self.buf = [Screen.Row(self) for i in range(self.ui.h)]

    def __getitem__(self, i):
        return self.buf[i]

    def output(self):
        print("\r\n".join(x.output() for x in self.buf), end="")



from .arn import ARN
from .base_control import Describer, ResourceLister


class ListenerResourceLister(ResourceLister):
    prefix = "listener_list"
    title = "Listeners"

    def title_info(self):
        return self.lb["name"]

    def __init__(self, *args, **kwargs):
        self.resource_key = "elbv2"
        self.list_method = "describe_listeners"
        self.item_path = ".Listeners"
        self.lb = kwargs["lb"]
        self.list_kwargs = {"LoadBalancerArn": self.lb["arn"]}
        self.column_paths = {
            "protocol": ".Protocol",
            "port": ".Port",
            "ssl policy": self.determine_ssl_policy,
        }
        self.imported_column_sizes = {
            "protocol": 10,
            "port": 10,
            "ssl policy": 30,
        }
        self.hidden_columns = {
            "arn": ".ListenerArn",
            "name": ".ListenerArn",
        }
        self.describe_command = ListenerDescriber.opener
        self.open_command = ListenerActionResourceLister.opener
        self.open_selection_arg = "listener"
        self.primary_key = "arn"

        self.imported_column_order = ["protocol", "port", "ssl policy"]
        self.sort_column = "arn"
        super().__init__(*args, **kwargs)

    def determine_ssl_policy(self, l, *args):
        if l["Protocol"] == "HTTPS":
            return l["SslPolicy"]
        return ""


class ListenerDescriber(Describer):
    prefix = "listener_browser"
    title = "Listener"

    def __init__(
        self, parent, alignment, dimensions, entry, *args, entry_key="name", **kwargs
    ):
        self.resource_key = "elbv2"
        self.describe_method = "describe_listeners"
        self.describe_kwarg_name = "ListenerArns"
        self.describe_kwarg_is_list = True
        self.object_path = ".Listeners[0]"
        super().__init__(
            parent,
            alignment,
            dimensions,
            *args,
            entry=entry,
            entry_key=entry_key,
            **kwargs
        )


class ListenerActionResourceLister(ResourceLister):
    prefix = "listener_action_list"
    title = "Listener Rules"

    def title_info(self):
        return self.listener["name"]

    def __init__(self, *args, **kwargs):
        from .resource_tg import TargetGroupResourceLister

        self.resource_key = "elbv2"
        self.list_method = "describe_rules"
        self.item_path = ".Rules"
        self.listener = kwargs["listener"]
        self.list_kwargs = {"ListenerArn": self.listener["arn"]}
        self.column_paths = {
            "prio": ".Priority",
            "condition": self.determine_condition,
            "action type": self.determine_action_type,
            "target": self.determine_target,
        }
        self.imported_column_sizes = {
            "prio": 10,
            "condition": 40,
            "action type": 20,
            "target": 120,
        }
        self.hidden_columns = {
            "name": ".RuleArn",
            "arn": ".RuleArn",
        }
        self.describe_command = ListenerActionDescriber.opener
        self.open_command = TargetGroupResourceLister.opener
        self.open_selection_arg = "rule_entry"

        self.imported_column_order = ["prio", "condition", "action type", "target"]
        self.sort_column = "prio"
        super().__init__(*args, **kwargs)

    def open(self, *args):
        if self.selection is not None:
            if self.selection["action type"] == "forward":
                return super().open(*args)
            else:
                return self.describe(*args)

    def determine_condition(self, la, *args):
        if len(la["Conditions"]) > 1:
            return "<multiple>"
        elif len(la["Conditions"]) == 1:
            cond = la["Conditions"][0]
            field = cond["Field"]
            if field == "host-header":
                if "Values" in cond and len(cond["Values"]):
                    if isinstance(cond["Values"], str):
                        return "Host: {0}".format(cond["Values"])
                    else:
                        src = cond["Values"]
                else:
                    src = cond["HostHeaderConfig"]["Values"]
                return "Host: {0}".format("|".join(src))
            elif field == "path":
                if "Values" in cond and len(cond["Values"]):
                    if isinstance(cond["Values"], str):
                        return "Path: {0}".format(cond["Values"])
                    else:
                        src = cond["Values"]
                else:
                    src = cond["PathPatternConfig"]["Values"]
                return "Path: {0}".format("|".join(src))

            return field
        else:
            return "<always>"

    def determine_action_type(self, l, *args):
        if len(l["Actions"]) > 0:
            act = l["Actions"][-1]
            return act["Type"]
        else:
            return "N/A"

    def determine_target(self, l, *args):
        if len(l["Actions"]) > 0:
            act = l["Actions"][-1]
            if act["Type"] == "forward":
                return ",".join(
                    [
                        ARN(i["TargetGroupArn"]).resource_id_first
                        for i in act["ForwardConfig"]["TargetGroups"]
                    ]
                )
            elif act["Type"] == "redirect":
                r = act["RedirectConfig"]
                proto = (
                    "http(s)"
                    if r["Protocol"] == "#{protocol}"
                    else r["Protocol"].lower()
                )
                proto_ports = []
                if proto in ["http", "http(s)"]:
                    proto_ports.append("80")
                if proto in ["https", "http(s)"]:
                    proto_ports.append("443")
                port = ":{0}".format(r["Port"]) if r["Port"] not in proto_ports else ""
                return "{4} {0}://{1}{2}{3}".format(
                    proto,
                    r["Host"],
                    port,
                    r["Path"],
                    "301" if r["StatusCode"] == "HTTP_301" else "302",
                )
        return ""


class ListenerActionDescriber(Describer):
    prefix = "listener_action_browser"
    title = "Listener Rule"

    def __init__(
        self, parent, alignment, dimensions, entry, *args, entry_key="arn", **kwargs
    ):
        self.resource_key = "elbv2"
        self.describe_method = "describe_rules"
        self.describe_kwarg_name = "RuleArns"
        self.describe_kwarg_is_list = True
        self.describe_kwargs = {"RuleArns": [self.rule_arn]}
        self.object_path = ".Rules[0]"
        super().__init__(
            parent,
            alignment,
            dimensions,
            *args,
            entry=entry,
            entry_key=entry_key,
            **kwargs
        )



from .common import Common
from .info import HotkeyDisplay
from .termui.alignment import Dimension, TopRightAnchor
from .termui.list_control import ListControl, ListEntry


class RegionList(ListControl):
    def __init__(self, parent, alignment, dimensions, *args, **kwargs):
        super().__init__(
            parent,
            alignment,
            dimensions,
            color=Common.color("region_list_generic", "generic"),
            selection_color=Common.color("region_list_selection", "selection"),
            title_color=Common.color("region_list_heading", "column_title"),
            *args,
            **kwargs,
        )
        self.hotkey_display = HotkeyDisplay(
            self.parent,
            TopRightAnchor(1, 0),
            Dimension("33%|50", 8),
            self,
            session=Common.Session,
            highlight_color=Common.color("hotkey_display_title"),
            generic_color=Common.color("hotkey_display_value"),
        )
        self.add_hotkey("d", self.set_default_region, "Set as default")
        self.add_hotkey("KEY_ENTER", self.select_region, "Select region")
        self.add_column("usage frequency", 12)
        self.add_column("default", 8)
        idx = 0
        defa = 0
        for region in sorted(Common.Session.service_provider.list_regions()):
            if region == Common.Configuration["default_region"]:
                defa = idx
                d = "✓"
            else:
                d = " "
            self.add_entry(ListEntry(region, **{"usage frequency": 0, "default": d}))
            idx += 1
        self.selected = defa

    def set_default_region(self, _):
        if self.selection is not None:
            Common.Configuration["default_region"] = self.selection.name
            Common.Configuration.write_config()
            for entry in self.entries:
                if entry.name != Common.Configuration["default_region"]:
                    entry.columns["default"] = " "
                else:
                    entry.columns["default"] = "✓"

    def select_region(self, _):
        if self.selection is not None:
            Common.Session.region = self.selection.name



import subprocess

from .base_control import Describer, ResourceLister
from .common import Common, SessionAwareDialog
from .termui.alignment import CenterAnchor, Dimension
from .termui.control import Border
from .termui.dialog import DialogFieldText


class RDSResourceLister(ResourceLister):
    prefix = "rds_list"
    title = "RDS Instances"
    command_palette = ["rds"]

    def title_info(self):
        return self.title_info_data

    def __init__(self, *args, **kwargs):
        self.resource_key = "rds"
        self.list_method = "describe_db_instances"
        self.title_info_data = None
        self.item_path = ".DBInstances"
        self.column_paths = {
            "instance id": ".DBInstanceIdentifier",
            "host": ".Endpoint.Address",
            "engine": ".Engine",
            "type": ".DBInstanceClass",
            "vpc": ".DBSubnetGroup.VpcId",
        }
        self.hidden_columns = {
            "public_access": ".PubliclyAccessible",
            "db_name": ".DBName",
            "name": self.tag_finder_generator("Name", taglist_key="TagList"),
        }
        self.imported_column_sizes = {
            "instance id": 11,
            "host": 45,
            "engine": 10,
            "type": 10,
            "vpc": 15,
        }
        self.describe_command = RDSDescriber.opener
        self.imported_column_order = ["instance id", "host", "engine", "type", "vpc"]
        self.sort_column = "instance id"
        self.primary_key = "instance id"
        super().__init__(*args, **kwargs)
        self.add_hotkey("s", self.db_client, "Open command line")

    def db_client(self, _):
        if self.selection is not None:
            if self.selection["public_access"] != "True":
                Common.Session.set_message(
                    "No public IP associated with instance",
                    Common.color("message_error"),
                )
            else:
                RDSClientDialog(
                    self.parent,
                    CenterAnchor(0, 0),
                    Dimension("80%|40", "10"),
                    instance_entry=self.selection,
                    caller=self,
                    weight=-500,
                )


class RDSDescriber(Describer):
    prefix = "rds_browser"
    title = "RDS Instance"

    def __init__(
        self,
        parent,
        alignment,
        dimensions,
        entry,
        *args,
        entry_key="instance id",
        **kwargs
    ):
        self.resource_key = "rds"
        self.describe_method = "describe_db_instances"
        self.describe_kwarg_name = "DBInstanceIdentifier"
        self.object_path = ".DBInstances[0]"
        super().__init__(
            parent,
            alignment,
            dimensions,
            *args,
            entry=entry,
            entry_key=entry_key,
            **kwargs
        )


class RDSClientDialog(SessionAwareDialog):
    def __init__(
        self,
        parent,
        alignment,
        dimensions,
        instance_entry=None,
        caller=None,
        *args,
        **kwargs
    ):
        kwargs["border"] = Border(
            Common.border("rds_client_modal", "default"),
            Common.color("rds_client_modal_dialog_border", "modal_dialog_border"),
            "SSH to instance",
            Common.color(
                "rds_client_modal_dialog_border_title", "modal_dialog_border_title"
            ),
            instance_entry["instance id"],
            Common.color(
                "rds_client_modal_dialog_border_title_info",
                "modal_dialog_border_title_info",
            ),
        )
        super().__init__(parent, alignment, dimensions, caller=caller, *args, **kwargs)
        self.instance_id = instance_entry["instance id"]
        self.db_name = instance_entry["db_name"]
        self.ip = instance_entry["host"]
        self.engine = instance_entry["engine"]
        self.username_textfield = DialogFieldText(
            "Username",
            text="",
            color=Common.color(
                "rds_client_modal_dialog_textfield", "modal_dialog_textfield"
            ),
            selected_color=Common.color(
                "rds_client_modal_dialog_textfield_selected",
                "modal_dialog_textfield_selected",
            ),
            label_color=Common.color(
                "rds_client_modal_dialog_textfield_label",
                "modal_dialog_textfield_label",
            ),
            label_min=16,
        )
        self.password_textfield = DialogFieldText(
            "Password",
            text="",
            color=Common.color(
                "rds_client_modal_dialog_textfield", "modal_dialog_textfield"
            ),
            selected_color=Common.color(
                "rds_client_modal_dialog_textfield_selected",
                "modal_dialog_textfield_selected",
            ),
            label_color=Common.color(
                "rds_client_modal_dialog_textfield_label",
                "modal_dialog_textfield_label",
            ),
            label_min=16,
            password=True,
        )
        self.database_textfield = DialogFieldText(
            "Database",
            text=self.db_name,
            color=Common.color(
                "rds_client_modal_dialog_textfield", "modal_dialog_textfield"
            ),
            selected_color=Common.color(
                "rds_client_modal_dialog_textfield_selected",
                "modal_dialog_textfield_selected",
            ),
            label_color=Common.color(
                "rds_client_modal_dialog_textfield_label",
                "modal_dialog_textfield_label",
            ),
            label_min=16,
        )
        self.add_field(self.username_textfield)
        self.add_field(self.password_textfield)
        self.add_field(self.database_textfield)
        self.highlighted = 1 if def_text != "" else 0
        self.caller = caller

    def accept_and_close(self):
        if self.engine in ["aurora", "aurora-mysql", "mariadb", "mysql"]:
            dollar_zero = "mysql"
            cmd = "mysql -h {0} -D {1} -u {2} --password={3}".format(
                self.ip,
                self.database_textfield.text,
                self.username_textfield.text,
                self.password_textfield.text,
            )
        elif self.engine in ["aurora-postgresql", "postgres"]:
            dollar_zero = "psql"
            cmd = "psql postgres://{2}:{3}@{0}/{1}".format(
                self.ip,
                self.database_textfield.text,
                self.username_textfield.text,
                self.password_textfield.text,
            )
        else:
            Common.Session.set_message(
                "Unsupported engine: {0}".format(self.engine),
                Common.color("message_info"),
            )
            self.close()
            return
        ex = Common.Session.ui.unraw(subprocess.run, ["bash", "-c", cmd])
        Common.Session.set_message(
            "{1} exited with code {0}".format(ex.returncode, dollar_zero),
            Common.color("message_info"),
        )
        super().accept_and_close()



from .base_control import Describer, ResourceLister


class AMIResourceLister(ResourceLister):
    prefix = "ami_list"
    title = "Amazon Machine Images"
    command_palette = ["ami", "image"]

    def title_info(self):
        return self.title_info_data

    def __init__(self, *args, **kwargs):
        self.resource_key = "ec2"
        self.list_method = "describe_images"
        self.title_info_data = None
        self.list_kwargs = {"Owners": ["self"]}
        if "ec2" in kwargs:
            self.list_kwargs["ImageIds"] = [kwargs["ec2"]["image"]]
            self.title_info_data = "Instance: {0}".format(kwargs["ec2"]["instance id"])
        self.item_path = ".Images"
        self.column_paths = {
            "id": ".ImageId",
            "name": ".Name",
            "arch": ".Architecture",
            "platform": ".PlatformDetails",
            "type": ".ImageType",
            "owner": ".ImageOwnerAlias",
            "state": ".State",
            "virt": ".VirtualizationType",
        }
        self.imported_column_sizes = {
            "id": 15,
            "name": 64,
            "arch": 8,
            "platform": 10,
            "type": 10,
            "owner": 15,
            "state": 10,
            "virt": 10,
        }
        self.describe_command = AMIDescriber.opener
        self.imported_column_order = [
            "id",
            "name",
            "arch",
            "platform",
            "type",
            "owner",
            "state",
            "virt",
        ]
        self.primary_key = "id"
        self.sort_column = "id"
        super().__init__(*args, **kwargs)


class AMIDescriber(Describer):
    prefix = "ami_browser"
    title = "Amazon Machine Image"

    def __init__(
        self, parent, alignment, dimensions, entry, *args, entry_key="id", **kwargs
    ):
        self.resource_key = "ec2"
        self.describe_method = "describe_images"
        self.describe_kwarg_name = "ImageIds"
        self.describe_kwarg_is_list = True
        self.object_path = ".Images[0]"
        super().__init__(
            parent,
            alignment,
            dimensions,
            *args,
            entry=entry,
            entry_key=entry_key,
            **kwargs
        )



from pathlib import Path

import yaml

from .scheme import Scheme
from .storage import Keystore

last_config_version = 12


class Config:
    def __init__(self, path=None):
        print("Initializing AWSC configuration...")
        self.version_updaters = {
            1: self.update_1,
            2: self.update_2,
            3: self.update_3,
            4: self.update_4,
            5: self.update_5,
            6: self.update_6,
            7: self.update_7,
            8: self.update_8,
            9: self.update_9,
            10: self.update_10,
            11: self.update_11,
            12: self.update_12,
        }
        if path is None:
            path = Path.home() / ".config" / "awsc"
        else:
            path = Path(path)
        if not path.exists():
            path.mkdir(parents=True, exist_ok=True)
        self.path = path
        self.keystore = Keystore(self)
        self.scheme = Scheme(self)

        self.config_path = self.path / "config.yaml"

        if not self.config_path.exists():
            self.create_default_config()
        else:
            self.parse_config()
        self.update_version()

    def update_1(self):
        self.config["default_region"] = "us-east-1"

    def update_2(self):
        self.scheme.backup_and_reset()

    def update_3(self):
        self.scheme.backup_and_reset()

    def update_4(self):
        self.scheme.backup_and_reset()
        self.config["usage_statistics"] = {
            "regions_by_context": {},
            "resources_by_context": {},
        }

    def update_5(self):
        self.scheme.backup_and_reset()

    def update_6(self):
        self.config["default_ssh_key"] = "id_rsa"

    def update_7(self):
        self.scheme.backup_and_reset()
        self.config["default_ssh_usernames"] = {}

    def update_8(self):
        self.config["editor_command"] = "nano {0}"

    def update_9(self):
        self.config["keypair_associations"] = {}

    def update_10(self):
        self.scheme.backup_and_reset()

    def update_11(self):
        self.scheme.backup_and_reset()

    def update_12(self):
        self.config["log_retention"] = {
            "max_lines": -1,
            "max_age": 2419200,
        }

    # TODO: Weigh whether it's worth implementing fshook (may be potentially dangerous)
    def update_X(self):
        self.config["editor_use_fshook"] = False

    def update_version(self):
        if "version" not in self.config:
            version = 0
        else:
            version = self.config["version"]
        while version < last_config_version:
            print("Performing config update to version {0}".format(version + 1))
            self.version_updaters[version + 1]()
            version += 1
        self.config["version"] = version
        self.write_config()

    def create_default_config(self):
        print("Creating first time configuration...")
        self.config = {
            "version": last_config_version,
            "contexts": {},
            "default_context": "",
            "default_region": "us-east-1",
            "default_ssh_key": "id_rsa",
            "default_ssh_usernames": {},
            "keypair_associations": {},
            "editor_command": "nano {0}",
            "log_retention": {
                "max_lines": -1,
                "max_age": 2419200,
            },
            "usage_statistics": {
                "regions": {},
                "resources": {},
                "keys": {},
            },
        }

        self.write_config()

    def write_config(self):
        with self.config_path.open("w") as f:
            f.write(yaml.dump(self.config))

    def __getitem__(self, item):
        return self.config[item]

    def __setitem__(self, item, value):
        self.config[item] = value

    def parse_config(self):
        with self.config_path.open("r") as f:
            self.config = yaml.safe_load(f.read())

    def add_or_edit_context(self, name, acctid, access, secret):
        self.config["contexts"][name] = {
            "account_id": acctid,
        }
        if self.config["default_context"] == "":
            self.config["default_context"] = name
        self.keystore.set_key(name, access, secret)
        self.write_config()

    def delete_context(self, name):
        del self.config["contexts"][name]
        if self.config["default_context"] == name:
            if len(self.config["contexts"]):
                self.config["default_context"] = self.config["contexts"].keys()[0]
        self.keystore.delete_key(name)
        self.write_config()



from .base_control import Describer, ResourceLister


class InstanceClassResourceLister(ResourceLister):
    prefix = "instance_class_list"
    title = "Instance Classes"
    command_palette = ["it", "instancetype", "instanceclass"]

    def __init__(self, *args, **kwargs):
        self.resource_key = "ec2"
        self.list_method = "describe_instance_types"
        self.item_path = ".InstanceTypes"
        self.list_kwargs = {"MaxResults": 20}
        self.column_paths = {
            "name": ".InstanceType",
            "cpus": ".VCpuInfo.DefaultVCpus",
            "memory": lambda x: "{0} MiB".format(x["MemoryInfo"]["SizeInMiB"]),
            "ebs optimization": ".EbsInfo.EbsOptimizedSupport",
            "network": ".NetworkInfo.NetworkPerformance",
        }
        self.imported_column_sizes = {
            "name": 20,
            "cpus": 4,
            "memory": 10,
            "ebs optimization": 15,
            "network": 15,
        }
        self.next_marker = "NextToken"
        self.next_marker_arg = "NextToken"
        self.imported_column_order = [
            "name",
            "cpus",
            "memory",
            "ebs optimization",
            "network",
        ]
        self.sort_column = "name"
        self.primary_key = "name"
        self.describe_command = InstanceClassDescriber.opener
        super().__init__(*args, **kwargs)


class InstanceClassDescriber(Describer):
    prefix = "instance_class_browser"
    title = "Instance Class"

    def __init__(
        self, parent, alignment, dimensions, entry, *args, entry_key="name", **kwargs
    ):
        self.resource_key = "ec2"
        self.describe_method = "describe_instance_types"
        self.describe_kwarg_name = "InstanceTypes"
        self.describe_kwarg_is_list = True
        self.object_path = ".InstanceTypes[0]"
        super().__init__(
            parent,
            alignment,
            dimensions,
            *args,
            entry=entry,
            entry_key=entry_key,
            **kwargs
        )



import configparser
import json
import sys
from datetime import datetime, timezone
from pathlib import Path
from typing import Any, Callable, Dict, Set, Union

import yaml
from botocore import exceptions

from .config.config import Config
from .session import Session
from .termui.alignment import CenterAnchor, Dimension, TopLeftAnchor
from .termui.bar_graph import BarGraph
from .termui.color import Color, Palette8Bit
from .termui.control import Border, BorderStyle
from .termui.dialog import DialogControl

DefaultAnchor = TopLeftAnchor(0, 11)
DefaultDimension = Dimension("100%", "100%-14")


def datetime_hack(x):
    if isinstance(x, datetime.datetime):
        return x.isoformat()
    raise TypeError("Unknown type")


class BaseChart(BarGraph):
    prefix = "CHANGEME"
    title = "CHANGEME"

    @classmethod
    def opener(cls, *args, **kwargs):
        ret = cls(
            Common.Session.ui.top_block,
            DefaultAnchor,
            DefaultDimension,
            *args,
            weight=0,
            color=Common.color("generic"),
            **kwargs,
        )
        ret.border = DefaultBorder(cls.prefix, cls.title, ret.title_info())
        return [ret]

    def title_info(self):
        return None


class SessionAwareDialog(DialogControl):
    @classmethod
    def opener(cls, caller, *args, **kwargs):
        return cls(
            caller.parent,
            CenterAnchor(0, 0),
            Dimension("80%|40", "10"),
            caller=caller,
            *args,
            **kwargs,
        )

    def __init__(self, *args, caller, **kwargs):
        caller.dialog_mode = True
        self.caller = caller
        Common.Session.extend_frame(self)
        kwargs["ok_action"] = self.accept_and_close
        kwargs["cancel_action"] = self.close
        super().__init__(*args, **kwargs)

    def input(self, key):
        if key.is_sequence and key.name == "KEY_ESCAPE":
            self.close()
            return True
        return super().input(key)

    def accept_and_close(self):
        self.close()

    def close(self):
        self.caller.dialog_mode = False
        self.parent.remove_block(self)
        Common.Session.remove_from_frame(self)


class Common:
    Configuration: Union[Config, Dict[Any, Any]] = {}
    Session = None
    _logholder = None
    initialized = False
    init_hooks: Set[Callable[[], None]] = set()

    @classmethod
    def run_on_init(cls, hook):
        if cls.initialized:
            hook()
        else:
            cls.init_hooks.add(hook)

    @classmethod
    def initialize(cls):
        cls.Configuration = Config()
        cls.Session = Session(
            cls.Configuration,
            cls.color("info_display_title"),
            cls.color("info_display_value"),
            cls,
        )
        cls._logholder = LogHolder()

    @classmethod
    def post_initialize(cls):
        for hook in cls.init_hooks:
            hook()
        cls.load_dot_aws()

    @classmethod
    def load_dot_aws(cls):
        aws_creds = Path.home() / ".aws" / "credentials"
        print("Loading ~/.aws/credentials", file=sys.stderr)
        try:
            with aws_creds.open("r") as f:
                creds = f.read()
        except OSError as e:
            print(
                "Failed to open ~/.aws/credentials: {0}".format(str(e)), file=sys.stderr
            )
            return
        parser = configparser.ConfigParser(default_section="__default")
        parser.read_string(creds)
        for section in parser.sections():
            if "aws_access_key_id" not in parser[section]:
                print(
                    "aws_access_key_id missing for credential {0}, skipping".format(
                        section
                    ),
                    file=sys.stderr,
                )
                continue
            if "aws_secret_access_key" not in parser[section]:
                print(
                    "aws_secret_access_key missing for credential {0}, skipping".format(
                        section
                    ),
                    file=sys.stderr,
                )
                continue
            access = parser[section]["aws_access_key_id"]
            secret = parser[section]["aws_secret_access_key"]
            api_keypair = {"access": access, "secret": secret}
            try:
                whoami = cls.Session.service_provider.whoami(keys=api_keypair)
            except exceptions.ClientError as e:
                cls.clienterror(
                    e,
                    "Verify Credentials",
                    "Bootstrap",
                    subcategory="Credentials Import",
                    resource=section,
                    set_message=False,
                    api_provider="sts",
                    api_method="get_caller_identity",
                    api_keypair=api_keypair,
                    api_args={},
                    credentials_section=section,
                )
                continue
            cls.Configuration.add_or_edit_context(
                section, whoami["Account"], access, secret
            )
            print(
                "Added {0} context from aws credentials file".format(section),
                file=sys.stderr,
            )

    @staticmethod
    def color(name, fallback=None):
        if Common.Configuration is None:
            raise ValueError("Configuration is not initialized.")
        if name not in Common.Configuration.scheme["colors"]:
            if fallback is None:
                raise KeyError('Undefined color "{0}"'.format(name))
            return Common.color(fallback)
        return Color(
            Palette8Bit(),
            Common.Configuration.scheme["colors"][name]["foreground"],
            background=Common.Configuration.scheme["colors"][name]["background"],
        )

    @staticmethod
    def border(name, fallback=None):
        if Common.Configuration is None:
            raise ValueError("Configuration is not initialized.")
        if name not in Common.Configuration.scheme["borders"]:
            if fallback is None:
                raise KeyError('Undefined border "{0}"'.format(name))
            return Common.border(fallback)
        border = Common.Configuration.scheme["borders"][name]
        return BorderStyle(
            [
                border["horizontal"],
                border["vertical"],
                border["TL"],
                border["TR"],
                border["BL"],
                border["BR"],
            ]
        )

    @staticmethod
    def main():
        Common.Session.ui.main()

    @staticmethod
    def confdir():
        return Path.home() / ".config" / "awsc"

    @classmethod
    def log(
        cls,
        message,
        summary,
        category,
        message_type,
        subcategory=None,
        resource=None,
        set_message=True,
        **kwargs,
    ):
        cls._logholder.log(
            message,
            summary,
            category,
            message_type,
            subcategory=subcategory,
            resource=resource,
            set_message=set_message,
            **kwargs,
        )

    @classmethod
    def error(
        cls,
        message,
        summary,
        category,
        subcategory=None,
        resource=None,
        set_message=True,
        **kwargs,
    ):
        cls.log(
            message,
            summary,
            category,
            "error",
            subcategory=subcategory,
            resource=resource,
            set_message=set_message,
            **kwargs,
        )

    @classmethod
    def clienterror(
        cls,
        error,
        summary,
        category,
        subcategory=None,
        resource=None,
        set_message=True,
        **kwargs,
    ):
        errtype = error.response["Error"]["Code"]
        errmsg = error.response["Error"]["Message"]
        cls.log(
            errmsg,
            "AWS: {0}".format(errtype),
            category,
            "error",
            subcategory=subcategory,
            resource=resource,
            set_message=set_message,
            **kwargs,
        )

    @classmethod
    def success(
        cls,
        message,
        summary,
        category,
        subcategory=None,
        resource=None,
        set_message=True,
        **kwargs,
    ):
        cls.log(
            message,
            summary,
            category,
            "success",
            subcategory=subcategory,
            resource=resource,
            set_message=set_message,
            **kwargs,
        )

    @classmethod
    def info(
        cls,
        message,
        summary,
        category,
        subcategory=None,
        resource=None,
        set_message=True,
        **kwargs,
    ):
        cls.log(
            message,
            summary,
            category,
            "info",
            subcategory=subcategory,
            resource=resource,
            set_message=set_message,
            **kwargs,
        )

    @classmethod
    def generic_api_call(
        cls,
        service,
        method,
        api_kwargs,
        summary,
        category,
        subcategory=None,
        success_template=None,
        resource=None,
        **kwargs,
    ):
        try:
            response = getattr(cls.Session.service_provider(service), method)(
                **api_kwargs
            )
            if success_template is not None:
                cls.success(
                    success_template.format(resource, resource=resource, **api_kwargs),
                    summary,
                    category,
                    subcategory=subcategory,
                    resource=resource,
                    set_message=True,
                    api_provider=service,
                    api_method=method,
                    api_args=api_kwargs,
                    **kwargs,
                )
            return {"Success": True, "Response": response}
        except exceptions.ClientError as e:
            cls.clienterror(
                e,
                summary,
                category,
                subcategory=subcategory,
                resource=resource,
                set_message=True,
                api_provider=service,
                api_method=method,
                api_args=api_kwargs,
                **kwargs,
            )
            return {"Success": False, "Response": e.response}
        # pylint: disable=broad-except # Anything else that may occur is irrelevant and should just get logged and thrown back as an error.
        except Exception as e:
            cls.error(
                str(e),
                summary,
                category,
                subcategory=subcategory,
                resource=resource,
                set_message=True,
                api_provider=service,
                api_method=method,
                api_args=api_kwargs,
                **kwargs,
            )
            return {
                "Success": False,
                "Response": {
                    "Error": {
                        "Code": "Python.{0}".format(type(e).__name__),
                        "Message": str(e),
                    }
                },
            }


def DefaultBorder(prefix, title, title_info=None):
    return Border(
        Common.border("resource_list", "default"),
        Common.color("{0}_border".format(prefix), "generic_border"),
        title,
        Common.color("{0}_border_title".format(prefix), "border_title"),
        title_info,
        Common.color("{0}_border_title_info".format(prefix), "border_title_info"),
    )


class LogHolder:
    def __init__(self):
        self.raw_entries = []
        self.control = None
        self.parse_log_messages()

    def parse_log_messages(self):
        file = Common.confdir() / "log.yaml"
        if file.exists():
            with file.open("r") as f:
                self.raw_entries = yaml.safe_load(f.read())
        else:
            self.raw_entries = []
        self.parse_raw_entries()

    def parse_raw_entries(self):
        if self.control is not None:
            self.control.entries.clear()
            for entry in self.raw_entries:
                self.control.add_raw_entry(entry)
            self.control.sort()

    def write_raw_entries(self):
        file = Common.confdir() / "log.yaml"
        limit = Common.Configuration["log_retention"]["max_lines"]
        if limit > 0:
            max_idx = limit
        else:
            max_idx = len(self.raw_entries)
        age_limit = Common.Configuration["log_retention"]["max_age"]
        now = datetime.now(timezone.utc).timestamp()
        if age_limit > 0:
            for idx in range(len(self.raw_entries)):
                if now - self.raw_entries[idx]["timestamp"] > age_limit:
                    max_idx = idx
                    break

        self.raw_entries = self.raw_entries[:max_idx]

        with file.open("w") as f:
            f.write(yaml.dump(self.raw_entries))

    def attach(self, control):
        self.control = control
        for entry in self.raw_entries:
            self.control.add_raw_entry(entry)

    def detach(self):
        self.control = None

    def add_raw_entry(self, entry):
        if self.control is not None:
            self.control.add_raw_entry(entry)
        self.write_raw_entries()

    def log(
        self,
        message,
        summary,
        category,
        message_type,
        subcategory=None,
        resource=None,
        set_message=True,
        **kwargs,
    ):
        if set_message:
            color = "message_info"
            if message_type == "success":
                color = "message_success"
            elif message_type == "error":
                color = "message_error"
            Common.Session.set_message(message, Common.color(color))
        self.raw_entries.insert(
            0,
            {
                "summary": summary,
                "category": category,
                "subcategory": subcategory,
                "type": message_type,
                "message": message,
                "resource": resource,
                "timestamp": datetime.now(timezone.utc).timestamp(),
                "context": json.dumps(kwargs, default=datetime_hack),
            },
        )
        self.add_raw_entry(self.raw_entries[0])



from .common import Commons


class Block:
    def __init__(
        self, parent, alignment, dimensions, *args, weight=0, tag="default", **kwargs
    ):
        self.blocks = []
        self.weight = weight
        if parent is not None:
            parent.add_block(self)
        self.parent = parent
        self.alignment = alignment
        self.dimensions = dimensions
        self.tag = tag
        Commons.UIInstance.dirty = True

    def reparent(self):
        self.parent.remove_block(self)
        self.parent.add_block(self)

    def topleft(self):
        return self.alignment.topleft(self.dimensions, self.parent)

    def bottomright(self):
        tl = self.alignment.topleft(self.dimensions, self.parent)
        return (tl[0] + self.dimensions[0], tl[1] + self.dimensions[1])

    @property
    def w(self):
        c = self.corners()
        return c[0][1] - c[0][0] + 1

    @property
    def h(self):
        c = self.corners()
        return c[1][1] - c[1][0] + 1

    @property
    def w_in(self):
        return self.w

    def corners(self):
        tl = self.topleft()
        br = (tl[0] + self.dimensions[0] - 1, tl[1] + self.dimensions[1] - 1)
        return ((tl[0], br[0]), (tl[1], br[1]))

    def before_paint(self):
        for block in reversed(self.blocks):
            block.before_paint()

    def paint(self):
        for block in reversed(self.blocks):
            block.paint()

    def add_block(self, block):
        Commons.UIInstance.dirty = True
        idx = 0
        for i in range(len(self.blocks)):
            blk = self.blocks[i]
            if blk.weight > block.weight:
                idx = i
                break
            i += 1
        block.parent = self
        self.blocks.insert(idx, block)

    def clear_blocks(self, tag=None):
        self.blocks = [
            block for block in self.blocks if tag is not None and block.tag != tag
        ]

    def remove_block(self, block):
        try:
            self.blocks.remove(block)
        except ValueError:
            pass

    def input(self, key):
        for block in self.blocks:
            if block.input(key):
                return True
        return False



import datetime
import json
import threading
import traceback
from typing import List

import pyperclip
from botocore import exceptions as botoerror

from .common import (
    Common,
    DefaultAnchor,
    DefaultBorder,
    DefaultDimension,
    SessionAwareDialog,
)
from .info import HotkeyDisplay
from .termui.alignment import CenterAnchor, Dimension, TopRightAnchor
from .termui.color import ColorBlackOnGold, ColorBlackOnOrange, ColorGold
from .termui.control import Border
from .termui.dialog import (
    DialogField,
    DialogFieldCheckbox,
    DialogFieldLabel,
    DialogFieldText,
)
from .termui.list_control import ListControl, ListEntry
from .termui.text_browser import TextBrowser
from .termui.ui import ControlCodes


def datetime_hack(x):
    if isinstance(x, datetime.datetime):
        return x.isoformat()
    raise TypeError("Unknown type")


class StopLoadingData(Exception):
    pass


class ResourceListerBase(ListControl):
    def __init__(self, *args, **kwargs):
        if not hasattr(self, "primary_key"):
            self.primary_key = None
        self.load_counter = 1
        self.dialog_mode = False
        self.closed = False
        if not hasattr(self, "next_marker"):
            self.next_marker = None
        if not hasattr(self, "next_marker_arg"):
            self.next_marker_arg = None
        self.auto_refresh_last = datetime.datetime.now()
        super().__init__(*args, **kwargs)

    def on_close(self):
        self.closed = True

    def auto_refresh(self):
        pass

    def asynch(self, fn, *args, clear=False, **kwargs):
        try:
            t = threading.Thread(
                target=self.async_inner,
                args=args,
                kwargs={**kwargs, "fn": fn, "clear": clear},
                daemon=True,
            )
            t.start()
        except Exception as e:  # pylint: disable=broad-except # Purpose of this function is extremely generic
            Common.Session.set_message(
                "Failed to start AWS query thread: {0}".format(str(e)),
                Common.color("message_error"),
            )

    def before_paint_critical(self):
        super().before_paint_critical()
        if "thread_error" in self.thread_share:
            Common.Session.set_message(
                self.thread_share["thread_error"], Common.color("message_error")
            )
            del self.thread_share["thread_error"]

    def async_inner(self, *args, fn, clear=False, **kwargs):
        try:
            self.mutex.acquire()
            try:
                if clear or (
                    not hasattr(self, "primary_key") or self.primary_key is None
                ):
                    self.thread_share["clear"] = True
            finally:
                self.mutex.release()

            for data in fn(*args, **kwargs):
                self.mutex.acquire()
                try:
                    self.thread_share["new_entries"].extend(data)
                finally:
                    self.mutex.release()

        except StopLoadingData:
            pass
        except Exception as e:  # pylint: disable=broad-except # Purpose of this function is extremely generic, needs a catch-all
            self.thread_share[
                "thread_error"
            ] = "Refresh thread execution failed: {0}: {1}".format(
                e.__class__.__name__, str(e)
            )
            Common.error(
                "Refresh thread execution failed: {0}: {1}\n{2}".format(
                    e.__class__.__name__, str(e), traceback.format_exc()
                ),
                "Refresh Thread Error",
                "Core",
                set_message=False,
            )
            Common.Session.set_message(
                "Refresh thread execution failed", Common.color("message_error")
            )
        self.mutex.acquire()
        try:
            self.thread_share["updating"] = False
            self.thread_share["finalize"] = True
        finally:
            self.mutex.release()

    def handle_finalization_critical(self):
        if hasattr(self, "primary_key") and self.primary_key is not None:
            ne = []
            for entry in self.entries:
                if entry[self.primary_key] in self.thread_share["acquired"]:
                    ne.append(entry)
            self.entries = ne[:]
            self.sort()

    def handle_new_entries_critical(self, entries):
        if not hasattr(self, "primary_key") or self.primary_key is None:
            super().handle_new_entries_critical(entries)
        else:
            for new in entries:
                found = False
                for old in self.entries:
                    if old[self.primary_key] == new[self.primary_key]:
                        old.mutate(new)
                        found = True
                        break
                if not found:
                    self.entries.append(new)
                self.thread_share["acquired"].append(new[self.primary_key])

    def get_data_generic(
        self,
        resource_key,
        list_method,
        list_kwargs,
        item_path,
        column_paths,
        hidden_columns,
        next_marker_name,
        next_marker_arg,
        *args
    ):
        try:
            provider = Common.Session.service_provider(resource_key)
        except KeyError:
            Common.error(
                "boto3 does not recognize provider",
                "Invalid provider",
                "Core",
                subcategory="ResourceListerBase",
                api_provider=self.resource_key,
                classname=type(self).__name__,
            )
            return
        if callable(list_kwargs):
            list_kwargs = list_kwargs()
        ret = []
        next_marker = None
        while next_marker != "":
            if self.closed:
                raise StopLoadingData
            it_list_kwargs = list_kwargs.copy()
            if next_marker is not None and next_marker_arg is not None:
                it_list_kwargs[next_marker_arg] = next_marker
            try:
                response = getattr(provider, list_method)(**it_list_kwargs)
            except botoerror.ClientError as e:
                Common.clienterror(
                    e,
                    "List Resources",
                    "Core",
                    subcategory="ResourceListerBase",
                    api_provider=resource_key,
                    api_method=list_method,
                    api_kwargs=it_list_kwargs,
                )
                # pylint: disable=raise-missing-from # StopLoadingData is a special exception to stop this generator from being used.
                raise StopLoadingData

            it = (
                Common.Session.jq(item_path)
                .input(text=json.dumps(response, default=datetime_hack))
                .first()
            )
            if it is None:
                Common.error(
                    "get_data_generic returned None",
                    "Get Data Generic returned None",
                    "Core",
                    set_message=False,
                    api_provider=resource_key,
                    api_method=list_method,
                    api_kwargs=list_kwargs,
                    item_path=item_path,
                    api_response=response,
                )
                return []

            for item in it:
                if self.closed:
                    raise StopLoadingData
                init = {}
                for column, path in {**column_paths, **hidden_columns}.items():
                    if callable(path):
                        init[column] = path(item)
                    else:
                        try:
                            init[column] = Common.Session.jq(path).input(item).first()
                        except StopIteration:
                            init[column] = ""
                le = ListEntry(**init)
                le.controller_data = item
                if self.matches(le, *args):
                    ret.append(le)
            if self.closed:
                raise StopLoadingData
            yield ret
            if self.closed:
                raise StopLoadingData
            self.load_counter += 1
            if next_marker_name is not None and next_marker_name in response:
                next_marker = response[next_marker_name]
                ret = []
            else:
                next_marker = ""

    def matches(self, list_entry, *args):
        return True

    def tag_finder_generator(self, tagname, default="", taglist_key="Tags"):
        def fn(e, *args):
            if taglist_key in e:
                for tag in e[taglist_key]:
                    if tag["Key"] == tagname:
                        return tag["Value"]
            return default

        return fn

    def empty(self, _):
        return ""


class DialogFieldResourceListSelector(DialogField):
    def __init__(
        self,
        selector_class,
        label,
        default="",
        color=ColorBlackOnOrange,
        selected_color=ColorBlackOnGold,
        label_color=ColorGold,
        label_min=0,
        primary_key=None,
    ):
        super().__init__()
        self.highlightable = True
        self.left = 0
        self.text = default
        self.label = label
        self.color = color
        self.label_color = label_color
        self.label_min = label_min
        self.selected_color = selected_color
        self.centered = True
        self.selector_class = selector_class
        self.primary_key = primary_key

    def selector_callback(self, entry):
        self.text = entry

    def input(self, inkey):
        if inkey.is_sequence:
            if inkey.name == "KEY_ENTER":
                kwa = {}
                if self.primary_key is not None:
                    kwa["primary_key"] = self.primary_key
                Common.Session.push_frame(
                    self.selector_class.selector(self.selector_callback, **kwa)
                )
                Common.Session.ui.dirty = True
                return True
            elif inkey.name == "KEY_BACKSPACE" or inkey.name == "KEY_DELETE":
                self.text = ""
                Common.Session.ui.dirty = True
                return True
        return False

    def paint(self, x0, x1, y, selected=False):
        x = x0
        Common.Session.ui.print(self.label, xy=(x, y), color=self.label_color)
        x += max(len(self.label) + 1, self.label_min)
        space = x1 - x + 1
        t = self.text + " ↲"
        if self.left >= len(t):
            self.left = 0
        text = t[
            self.left : (self.left + space if len(t) > self.left + space else len(t))
        ]
        Common.Session.ui.print(
            text, xy=(x, y), color=self.selected_color if selected else self.color
        )


class SingleNameDialog(SessionAwareDialog):
    def __init__(
        self,
        parent,
        title,
        callback,
        *args,
        label="Name:",
        what="name",
        subject="",
        default="",
        caller=None,
        accepted_inputs=None,
        **kwargs
    ):
        kwargs["border"] = Border(
            Common.border("default"),
            Common.color("modal_dialog_border"),
            title,
            Common.color("modal_dialog_border_title"),
        )
        super().__init__(
            parent,
            CenterAnchor(0, 0),
            Dimension("80%|40", "10"),
            caller=caller,
            *args,
            **kwargs,
        )
        self.what = what
        self.add_field(
            DialogFieldLabel(
                "Enter {0}{1}".format(
                    what, " for {0}".format(subject) if subject != "" else ""
                )
            )
        )
        self.error_label = DialogFieldLabel(
            "", default_color=Common.color("modal_dialog_error")
        )
        self.add_field(self.error_label)
        self.add_field(DialogFieldLabel(""))
        self.name_field = DialogFieldText(
            label,
            label_min=16,
            color=Common.color("modal_dialog_textfield"),
            selected_color=Common.color("modal_dialog_textfield_selected"),
            label_color=Common.color("modal_dialog_textfield_label"),
            accepted_inputs=accepted_inputs,
        )
        self.add_field(self.name_field)
        self.caller = caller
        self.callback = callback

    def accept_and_close(self):
        if self.name_field.text == "":
            self.error_label.text = "You must enter a {0}.".format(self.what)
            return
        self.callback(self.name_field.text)
        super().accept_and_close()

    def close(self):
        if self.caller is not None:
            self.caller.refresh_data()
        super().close()


class SingleSelectorDialog(SessionAwareDialog):
    def __init__(
        self,
        parent,
        title,
        resource_type,
        action_name,
        selector_class,
        callback,
        *args,
        caller=None,
        **kwargs
    ):
        kwargs["border"] = Border(
            Common.border("default"),
            Common.color("modal_dialog_border"),
            title,
            Common.color("modal_dialog_border_title"),
        )
        super().__init__(
            parent,
            CenterAnchor(0, 0),
            Dimension("80%|40", "10"),
            *args,
            caller=caller,
            **kwargs,
        )
        self.add_field(
            DialogFieldLabel(
                "Select {0} for {1} action".format(resource_type, action_name)
            )
        )
        self.error_label = DialogFieldLabel(
            "", default_color=Common.color("modal_dialog_error")
        )
        self.add_field(self.error_label)
        self.add_field(DialogFieldLabel(""))
        self.resource_selector = DialogFieldResourceListSelector(
            selector_class,
            "{0}: ".format(resource_type),
            "",
            label_min=16,
            color=Common.color("modal_dialog_textfield"),
            selected_color=Common.color("modal_dialog_textfield_selected"),
            label_color=Common.color("modal_dialog_textfield_label"),
        )
        self.add_field(self.resource_selector)
        self.caller = caller
        self.callback = callback

    def accept_and_close(self):
        if self.resource_selector.text == "":
            self.error_label.text = "You must select a resource."
            return
        self.callback(self.resource_selector.text)
        super().accept_and_close()

    def close(self):
        if self.caller is not None:
            self.caller.refresh_data()
        super().close()


class ResourceLister(ResourceListerBase):
    prefix = "CHANGEME"
    title = "CHANGEME"
    command_palette: List[str] = []

    @classmethod
    def register(cls):
        for cmd in cls.command_palette:
            Common.Session.commander_options[cmd] = cls.opener
        for subcls in cls.__subclasses__():
            subcls.register()

    @classmethod
    def opener(cls, **kwargs):
        l = cls(
            Common.Session.ui.top_block,
            DefaultAnchor,
            DefaultDimension,
            weight=0,
            color=Common.color("{0}_generic".format(cls.prefix), "generic"),
            selection_color=Common.color(
                "{0}_selection".format(cls.prefix), "selection"
            ),
            title_color=Common.color("{0}_heading".format(cls.prefix), "column_title"),
            update_color=Common.color("{0}_updated".format(cls.prefix), "highlight"),
            update_selection_color=Common.color(
                "{0}_updated".format(cls.prefix), "highlight_selection"
            ),
            **kwargs,
        )
        l.border = DefaultBorder(cls.prefix, cls.title, l.title_info())
        return [l, l.hotkey_display]

    @classmethod
    def selector(cls, cb, **kwargs):
        return cls.opener(**{"selector_cb": cb, **kwargs})

    def title_info(self):
        return None

    def __init__(
        self, parent, alignment, dimensions, *args, selector_cb=None, **kwargs
    ):
        super().__init__(parent, alignment, dimensions, *args, **kwargs)
        if not hasattr(self, "resource_key"):
            raise AttributeError("resource_key is undefined")
        if not hasattr(self, "list_method"):
            raise AttributeError("list_method is undefined")
        if not hasattr(self, "list_kwargs"):
            if "list_kwargs" in kwargs:
                self.list_kwargs = kwargs["list_kwargs"]
            else:
                self.list_kwargs = {}
        if not hasattr(self, "describe_kwargs"):
            self.describe_kwargs = {}
        if not hasattr(self, "describe_command"):
            self.describe_command = None
        else:
            if not hasattr(self, "describe_selection_arg"):
                self.describe_selection_arg = "entry"
        if not hasattr(self, "open_command") or self.open_command is None:
            self.open_command = None
        else:
            if not isinstance(self.open_command, str):
                if not hasattr(self, "open_selection_arg"):
                    raise AttributeError(
                        "open_command is defined but open_selection_arg is undefined"
                    )
            elif (
                not hasattr(self, "additional_commands")
                or self.open_command not in self.additional_commands
            ):
                raise AttributeError(
                    "open_command refers to a key that is not in additional_commands"
                )
        if not hasattr(self, "item_path"):
            raise AttributeError("item_path is undefined")
        if not hasattr(self, "hidden_columns"):
            self.hidden_columns = {}
        if not hasattr(self, "column_paths"):
            raise AttributeError("column_paths is undefined")

        if "name" not in self.column_paths and "name" not in self.hidden_columns:
            raise AttributeError(
                "name entry is required in column_paths or hidden_columns"
            )
        if not hasattr(self, "imported_column_sizes"):
            self.imported_column_sizes = {
                k: len(k) for k in self.column_paths.keys() if k != "name"
            }
        if not hasattr(self, "imported_column_order"):
            self.imported_column_order = sorted(
                [k for k in self.column_paths.keys() if k != "name"]
            )
        if not hasattr(self, "sort_column"):
            self.sort_column = "name"
        if not hasattr(self, "primary_key") or self.primary_key is None:
            self.primary_key = "name"
        if "primary_key" in kwargs:
            self.primary_key = kwargs["primary_key"]
        if "update_color" in kwargs:
            self.update_color = kwargs["update_color"]
        else:
            self.update_color = self.color
        if "update_selection_color" in kwargs:
            self.update_selection_color = kwargs["update_selection_color"]
        else:
            self.update_selection_color = self.selection_color

        self.selector_cb = selector_cb

        if self.open_command is not None:
            open_tooltip = "Open"
            if isinstance(self.open_command, str):
                cmd = self.additional_commands[self.open_command]
                self.open_command = cmd["command"]
                self.open_selection_arg = cmd["selection_arg"]
                open_tooltip = cmd["tooltip"]
            if selector_cb is None:
                self.add_hotkey("KEY_ENTER", self.open, open_tooltip)
            else:
                self.add_hotkey("o", self.open, open_tooltip)
        if self.describe_command is not None:
            describe_tooltip = "Describe"
            if isinstance(self.describe_command, str):
                cmd = self.additional_commands[self.describe_command]
                self.describe_command = cmd["command"]
                self.describe_selection_arg = cmd["selection_arg"]
                describe_tooltip = cmd["tooltip"]
            self.add_hotkey("d", self.describe, "Describe")
            if self.open_command is None and selector_cb is None:
                self.add_hotkey("KEY_ENTER", self.describe, describe_tooltip)
        if selector_cb is not None:
            self.add_hotkey("KEY_ENTER", self.select_and_close, "Select")
        if hasattr(self, "additional_commands"):
            for key, command_spec in self.additional_commands.items():
                if "kwargs" not in command_spec:
                    command_spec["kwargs"] = {}
                self.add_hotkey(
                    key,
                    self.command_wrapper(
                        command_spec["command"],
                        command_spec["selection_arg"],
                        **command_spec["kwargs"],
                    ),
                    command_spec["tooltip"],
                )
        self.add_hotkey(ControlCodes.R, self.refresh_data, "Refresh")
        if "arn" in self.column_paths or "arn" in self.hidden_columns:
            self.add_hotkey("r", self.copy_arn, "Copy ARN")
        self.hotkey_display = HotkeyDisplay(
            self.parent,
            TopRightAnchor(1, 0),
            Dimension("33%|50", 8),
            self,
            session=Common.Session,
            highlight_color=Common.color("hotkey_display_title"),
            generic_color=Common.color("hotkey_display_value"),
        )

        if "name" in self.imported_column_sizes or "name" in self.hidden_columns:
            self.column_titles = {}
        self.column_titles.update(self.imported_column_sizes)
        if "name" in self.imported_column_order or "name" in self.hidden_columns:
            self.column_order = []
        self.column_order.extend(self.imported_column_order)
        self.refresh_data()

    def copy_arn(self, *args):
        if self.selection is not None:
            pyperclip.copy(self.selection["arn"])
            Common.Session.set_message(
                "Copied resource ARN to clipboard", Common.color("message_success")
            )

    def select_and_close(self, *args):
        import sys

        print("self.selection is {0}".format(self.selection), file=sys.stderr)
        print("self.selector_cb is {0}".format(self.selector_cb), file=sys.stderr)
        print("self.primary_key is {0}".format(self.primary_key), file=sys.stderr)
        if (
            self.selection is not None
            and self.selector_cb is not None
            and self.primary_key is not None
        ):
            self.selector_cb(self.selection[self.primary_key])
            Common.Session.pop_frame()

    def command(self, cmd, kw={}):
        if self.selection is not None:
            frame = cmd(**kw)
            if frame is not None:
                Common.Session.push_frame(frame)

    def command_wrapper(self, cmd, selection_arg, **kwargs):
        def fn(*args):
            self.command(
                cmd,
                {
                    selection_arg: self.selection,
                    "pushed": True,
                    "caller": self,
                    **kwargs,
                },
            )

        return fn

    def describe(self, *args):
        if self.describe_command is not None:
            self.command_wrapper(
                self.describe_command,
                self.describe_selection_arg,
                **self.describe_kwargs,
            )()

    def open(self, *args):
        if self.open_command is not None:
            self.command_wrapper(self.open_command, self.open_selection_arg)()

    def get_data(self, *args, **kwargs):
        for y in self.get_data_generic(
            self.resource_key,
            self.list_method,
            self.list_kwargs,
            self.item_path,
            self.column_paths,
            self.hidden_columns,
            self.next_marker,
            self.next_marker_arg,
        ):
            yield y

    def auto_refresh(self):
        if self.dialog_mode or self.primary_key is None:
            return
        if datetime.datetime.now() - self.auto_refresh_last > datetime.timedelta(
            seconds=10
        ):
            self.refresh_data()

    def refresh_data(self, *args, **kwargs):
        self.mutex.acquire()
        try:
            if "updating" in self.thread_share and self.thread_share["updating"]:
                return
            self.thread_share["updating"] = True
            self.thread_share["acquired"] = []
        finally:
            self.mutex.release()
        self.auto_refresh_last = datetime.datetime.now()
        self.asynch(self.get_data)

    def sort(self):
        self.entries.sort(key=lambda x: x.columns[self.sort_column])
        self._cache = None


class NoResults(Exception):
    pass


class SingleRelationLister(ResourceListerBase):
    prefix = "CHANGEME"
    title = "CHANGEME"

    @classmethod
    def opener(cls, **kwargs):
        l = cls(
            Common.Session.ui.top_block,
            DefaultAnchor,
            DefaultDimension,
            weight=0,
            color=Common.color("{0}_generic".format(cls.prefix), "generic"),
            selection_color=Common.color(
                "{0}_selection".format(cls.prefix), "selection"
            ),
            title_color=Common.color("{0}_heading".format(cls.prefix), "column_title"),
            **kwargs,
        )
        l.border = DefaultBorder(cls.prefix, cls.title, l.title_info())
        return [l, l.hotkey_display]

    def __init__(self, parent, alignment, dimensions, *args, **kwargs):
        super().__init__(parent, alignment, dimensions, *args, **kwargs)
        self.column_order = []
        self.column_titles = {}
        self.add_column("type", 30, 0)
        self.add_column("id", 120, 1)
        if not hasattr(self, "resource_key"):
            raise AttributeError("resource_key is undefined")
        if not hasattr(self, "describe_method"):
            raise AttributeError("describe_method is undefined")
        if not hasattr(self, "describe_kwargs"):
            raise AttributeError("describe_kwargs is undefined")
        if not hasattr(self, "object_path"):
            raise AttributeError("object_path is undefined")
        if not hasattr(self, "sort_column"):
            self.sort_column = "type"
        if not hasattr(self, "resource_descriptors"):
            raise AttributeError("resource_descriptors is undefined")
        self.descriptor = None
        self.descriptor_raw = None
        self.add_hotkey("d", self.describe, "Describe")
        self.add_hotkey("KEY_ENTER", self.describe, "Describe")
        self.hotkey_display = HotkeyDisplay(
            self.parent,
            TopRightAnchor(1, 0),
            Dimension("33%|50", 8),
            self,
            session=Common.Session,
            highlight_color=Common.color("hotkey_display_title"),
            generic_color=Common.color("hotkey_display_value"),
        )
        self.refresh_data()

    def describe(self, _):
        if self.selection is not None:
            if "describer" not in self.selection.controller_data:
                Common.Session.set_message(
                    "Resource cannot be described", Common.color("message_info")
                )
                return
            Common.Session.push_frame(
                self.selection.controller_data["describer"](
                    entry=self.selection, entry_key="id"
                )
            )

    def get_data(self, *args, **kwargs):
        if self.descriptor is None:
            try:
                provider = Common.Session.service_provider(self.resource_key)
            except KeyError:
                Common.error(
                    "boto3 does not recognize provider",
                    "Invalid provider",
                    "Core",
                    subcategory="SingleRelationLister",
                    api_provider=self.resource_key,
                    classname=type(self).__name__,
                )
                return
            try:
                resp = getattr(provider, self.describe_method)(**self.describe_kwargs)
            except botoerror.ClientError as e:
                Common.clienterror(
                    e,
                    "List Resources",
                    "Core",
                    subcategory="SingleRelationLister",
                    api_provider=self.resource_key,
                    api_method=self.describe_method,
                    api_kwargs=self.describe_kwargs,
                )
                return
            self.descriptor = (
                Common.Session.jq(self.object_path)
                .input(text=json.dumps(resp, default=datetime_hack))
                .first()
            )
            self.descriptor_raw = json.dumps(self.descriptor, default=datetime_hack)
        for elem in self.resource_descriptors:
            try:
                result = (
                    Common.Session.jq(elem["base_path"])
                    .input(text=self.descriptor_raw)
                    .first()
                )
                yield [
                    ListEntry(item, id=item, type=elem["type"], controller_data=elem)
                    for item in result
                ]
            except StopIteration:
                continue
            except ValueError:
                continue

    def refresh_data(self, *args, **kwargs):
        self.asynch(self.get_data, clear=True)

    def sort(self):
        self.entries.sort(key=lambda x: x.columns[self.sort_column])
        self._cache = None


class MultiLister(ResourceListerBase):
    prefix = "CHANGEME"
    title = "CHANGEME"

    @classmethod
    def opener(cls, **kwargs):
        l = cls(
            Common.Session.ui.top_block,
            DefaultAnchor,
            DefaultDimension,
            weight=0,
            color=Common.color("{0}_generic".format(cls.prefix), "generic"),
            selection_color=Common.color(
                "{0}_selection".format(cls.prefix), "selection"
            ),
            title_color=Common.color("{0}_heading".format(cls.prefix), "column_title"),
            **kwargs,
        )
        l.border = DefaultBorder(cls.prefix, cls.title, l.title_info())
        return [l, l.hotkey_display]

    def __init__(
        self,
        parent,
        alignment,
        dimensions,
        compare_value,
        *args,
        compare_key="id",
        **kwargs
    ):
        super().__init__(parent, alignment, dimensions, *args, **kwargs)
        self.add_column("type", 30, 0)
        self.add_column("id", 30, 1)
        if not hasattr(self, "resource_descriptors"):
            raise AttributeError("resource_descriptors is undefined")
        if isinstance(compare_value, ListEntry):
            self.compare_value = compare_value[compare_key]
            self.orig_compare_value = compare_value
        else:
            self.compare_value = compare_value
            self.orig_compare_value = compare_value
        self.add_hotkey("KEY_ENTER", self.describe, "Describe")
        self.add_hotkey("d", self.describe, "Describe")
        self.add_hotkey(ControlCodes.R, self.refresh_data, "Refresh")
        self.hotkey_display = HotkeyDisplay(
            self.parent,
            TopRightAnchor(1, 0),
            Dimension("33%|50", 8),
            self,
            session=Common.Session,
            highlight_color=Common.color("hotkey_display_title"),
            generic_color=Common.color("hotkey_display_value"),
        )
        self.refresh_data()

    def describe(self, *args):
        if self.selection is not None:
            Common.Session.push_frame(
                GenericDescriber.opener(
                    **{
                        "describing": self.selection["id"],
                        "content": json.dumps(
                            self.selection.controller_data, sort_keys=True, indent=2
                        ),
                        "pushed": True,
                    }
                )
            )

    def get_data(self, *args, **kwargs):
        for elem in self.resource_descriptors:
            try:
                for y in self.get_data_generic(
                    elem["resource_key"],
                    elem["list_method"],
                    elem["list_kwargs"],
                    elem["item_path"],
                    elem["column_paths"],
                    elem["hidden_columns"],
                    None,
                    None,
                    elem,
                ):
                    yield y
            except NoResults:
                continue

    def matches(self, list_entry, elem, *args):
        raw_item = list_entry.controller_data
        if "compare_as_list" not in elem or not elem["compare_as_list"]:
            if callable(elem["compare_path"]):
                val = elem["compare_path"](raw_item)
            else:
                try:
                    val = (
                        Common.Session.jq(elem["compare_path"]).input(raw_item).first()
                    )
                except ValueError as e:
                    return False
                except StopIteration as e:
                    return False
            return val == self.compare_value
        else:
            if callable(elem["compare_path"]):
                return self.compare_value in elem["compare_path"](raw_item)
            else:
                try:
                    for val in (
                        Common.Session.jq(elem["compare_path"]).input(raw_item).first()
                    ):
                        if val == self.compare_value:
                            return True
                except StopIteration as e:
                    return False
        return False

    def refresh_data(self, *args, **kwargs):
        self.asynch(self.get_data, clear=True)

    def sort(self):
        self.entries.sort(key=lambda x: x.columns["type"])
        self._cache = None


class GenericDescriber(TextBrowser):
    prefix = "generic_describer"
    title = "Describe resource"

    @classmethod
    def opener(cls, **kwargs):
        l = cls(
            Common.Session.ui.top_block,
            DefaultAnchor,
            DefaultDimension,
            weight=0,
            color=Common.color("{0}_generic".format(cls.prefix), "generic"),
            filtered_color=Common.color("{0}_filtered".format(cls.prefix), "selection"),
            **kwargs,
        )
        l.border = DefaultBorder(cls.prefix, cls.title, l.title_info())
        return [l, l.hotkey_display]

    def title_info(self):
        return self.describing

    def __init__(
        self, parent, alignment, dimensions, describing, content, *args, **kwargs
    ):
        super().__init__(parent, alignment, dimensions, *args, **kwargs)
        self.describing = describing
        self.add_text(content)
        self.hotkey_display = HotkeyDisplay(
            self.parent,
            TopRightAnchor(1, 0),
            Dimension("33%|50", 8),
            self,
            session=Common.Session,
            highlight_color=Common.color("hotkey_display_title"),
            generic_color=Common.color("hotkey_display_value"),
        )

    def toggle_wrap(self, *args, **kwargs):
        super().toggle_wrap(*args, **kwargs)
        Common.Session.set_message(
            "Text wrap {0}".format("ON" if self.wrap else "OFF"),
            Common.color("message_info"),
        )


class Describer(TextBrowser):
    prefix = "CHANGEME"
    title = "CHANGEME"

    @classmethod
    def opener(cls, **kwargs):
        l = cls(
            Common.Session.ui.top_block,
            DefaultAnchor,
            DefaultDimension,
            weight=0,
            color=Common.color("{0}_generic".format(cls.prefix), "generic"),
            filtered_color=Common.color("{0}_filtered".format(cls.prefix), "selection"),
            syntax_highlighting=True,
            scheme=Common.Configuration.scheme,
            **kwargs,
        )
        l.border = DefaultBorder(cls.prefix, cls.title, l.title_info())
        return [l, l.hotkey_display]

    def title_info(self):
        return self.entry_id

    def populate_entry(self, *args, entry, entry_key, **kwargs):
        if entry is None:
            self.entry = None
            self.entry_id = None
            return
        self.entry = entry
        self.entry_id = entry[entry_key]

    def populate_describe_kwargs(self):
        self.describe_kwargs[self.describe_kwarg_name] = (
            [self.entry_id] if self.describe_kwarg_is_list else self.entry_id
        )

    def __init__(
        self, parent, alignment, dimensions, *args, entry, entry_key, **kwargs
    ):
        self.populate_entry(entry=entry, entry_key=entry_key)
        super().__init__(parent, alignment, dimensions, *args, **kwargs)
        if not hasattr(self, "resource_key"):
            raise AttributeError("resource_key is undefined")
        if not hasattr(self, "describe_method"):
            raise AttributeError("describe_method is undefined")
        if not hasattr(self, "describe_kwarg_name") and not hasattr(
            self, "describe_kwargs"
        ):
            raise AttributeError(
                "describe_kwarg_name is undefined and describe_kwargs not set manually"
            )
        if not hasattr(self, "describe_kwarg_is_list"):
            self.describe_kwarg_is_list = False
        if not hasattr(self, "describe_kwargs"):
            self.describe_kwargs = {}
        if not hasattr(self, "object_path"):
            raise AttributeError("object_path is undefined")
        if hasattr(self, "describe_kwarg_name"):
            self.populate_describe_kwargs()

        self.add_hotkey(ControlCodes.R, self.refresh_data, "Refresh")
        self.hotkey_display = HotkeyDisplay(
            self.parent,
            TopRightAnchor(1, 0),
            Dimension("33%|50", 8),
            self,
            session=Common.Session,
            highlight_color=Common.color("hotkey_display_title"),
            generic_color=Common.color("hotkey_display_value"),
        )

        if hasattr(self, "additional_commands"):
            for key, command_spec in self.additional_commands.items():
                if "kwargs" not in command_spec:
                    command_spec[kwargs] = {}
                self.add_hotkey(
                    key,
                    self.command_wrapper(
                        command_spec["command"],
                        command_spec["data_arg"],
                        **command_spec["kwargs"],
                    ),
                    command_spec["tooltip"],
                )

        self.refresh_data()

    def command(self, cmd, kw={}):
        frame = cmd(**kw)
        if frame is not None:
            Common.Session.push_frame(frame)

    def command_wrapper(self, cmd, data_arg, **kwargs):
        def fn(*args):
            self.command(
                cmd,
                {
                    data_arg: "\n".join(self.lines),
                    "pushed": True,
                    "caller": self,
                    **kwargs,
                },
            )

        return fn

    def toggle_wrap(self, *args, **kwargs):
        super().toggle_wrap(*args, **kwargs)
        Common.Session.set_message(
            "Text wrap {0}".format("ON" if self.wrap else "OFF"),
            Common.color("message_info"),
        )

    def refresh_data(self, *args, **kwargs):
        try:
            provider = Common.Session.service_provider(self.resource_key)
        except KeyError:
            Common.error(
                "boto3 does not recognize provider",
                "Invalid provider",
                "Core",
                subcategory="Describer",
                api_provider=self.resource_key,
                classname=type(self).__name__,
            )
            return
        try:
            response = getattr(provider, self.describe_method)(**self.describe_kwargs)
        except botoerror.ClientError as e:
            Common.clienterror(
                e,
                "List Resources",
                "Core",
                subcategory="Describer",
                api_provider=self.resource_key,
                api_method=self.describe_method,
                api_kwargs=self.describe_kwargs,
            )
            return
        self.clear()
        self.add_text(
            json.dumps(
                Common.Session.jq(self.object_path)
                .input(text=json.dumps(response, default=datetime_hack))
                .first(),
                sort_keys=True,
                indent=2,
            )
        )


class DeleteResourceDialog(SessionAwareDialog):
    def __init__(
        self,
        parent,
        alignment,
        dimensions,
        *args,
        resource_type,
        resource_identifier,
        callback,
        action_name="Delete",
        from_what=None,
        from_what_name=None,
        undoable=False,
        can_force=False,
        extra_fields={},
        **kwargs
    ):
        kwargs["ok_action"] = self.accept_and_close
        kwargs["cancel_action"] = self.close
        kwargs["border"] = Border(
            Common.border("default"),
            Common.color("modal_dialog_border"),
            "{1} {0}".format(resource_type, action_name),
            Common.color("modal_dialog_border_title"),
        )
        super().__init__(parent, alignment, dimensions, *args, **kwargs)
        label = [
            ("{0} ".format(action_name), Common.color("modal_dialog_label")),
            (resource_type, Common.color("modal_dialog_label_highlight")),
            (' resource "', Common.color("modal_dialog_label")),
            (resource_identifier, Common.color("modal_dialog_label_highlight")),
            ('"', Common.color("modal_dialog_label")),
        ]

        if from_what is not None:
            label.append(
                (" from {0}".format(from_what), Common.color("modal_dialog_label"))
            )
            if from_what_name is not None:
                label.append((' "', Common.color("modal_dialog_label")))
                label.append(
                    (from_what_name, Common.color("modal_dialog_label_highlight"))
                )
                label.append(('"', Common.color("modal_dialog_label")))

        label.append(("?", Common.color("modal_dialog_label")))
        self.add_field(DialogFieldLabel(label))
        if not undoable:
            self.add_field(
                DialogFieldLabel(
                    "This action cannot be undone.", Common.color("modal_dialog_error")
                )
            )
        self.highlighted = 1
        self.can_force = can_force
        if self.can_force:
            self.highlighted += 1
            self.force_checkbox = DialogFieldCheckbox(
                "Force",
                checked=False,
                color=Common.color("modal_dialog_textfield_label"),
                selected_color=Common.color("modal_dialog_textfield_selected"),
            )
            self.add_field(self.force_checkbox)
        self.extra_fields = {}
        for field in extra_fields.keys():
            self.highlighted += 1
            self.add_field(extra_fields[field])
            self.extra_fields[field] = extra_fields
        self.callback = callback

    def input(self, inkey):
        if inkey.is_sequence and inkey.name == "KEY_ESCAPE":
            self.close()
            return True
        return super().input(inkey)

    def accept_and_close(self):
        force = self.can_force and self.force_checkbox.checked
        self.callback(force=force, **self.extra_fields)
        self.close()


def format_timedelta(delta):
    hours = int(delta.seconds / 3600)
    minutes = int(delta.seconds / 60) - hours * 60
    seconds = delta.seconds - hours * 3600 - minutes * 60
    if delta.days > 0:
        return "{0}d{1}h ago".format(delta.days, hours)
    elif hours > 0:
        return "{0}h{1}m ago".format(hours, minutes)
    elif minutes > 0:
        return "{0}m{1}s ago".format(minutes, seconds)
    elif seconds > 0:
        return "{0}s ago".format(seconds)
    else:
        return "<1s ago"


def isdumpable(v):
    return not (
        isinstance(v, int)
        or isinstance(v, float)
        or isinstance(v, str)
        or isinstance(v, bool)
    )


class ListResourceDocumentEditor:
    def __init__(
        self,
        provider,
        retrieve_method,
        retrieve_path,
        update_method,
        entry_name_arg,
        update_document_arg,
        entry_key="name",
        entry_name_arg_update=None,
        as_json=True,
        message="Update successful",
    ):
        """
        Initializes a ListResourceDocumentEditor.

            Parameters:
                provider (str): boto3 client provider name
                retrieve_method (str): The method name for the provider to retrieve the document of the edited resource
                retrieve_path (str): Path to the root of the document in the response for the retrieve method
                update_method (str): The method name for the provider to call to update the resource
                entry_name_arg (str): The name of the kwarg for the retrieve_method which must be passed the entry's key.
                update_document_arg (str): The name of the kwarg for the update_method which receives the updated document.
                entry_key (str): The field in the entry which contains the value for the entry name argument.
                entry_name_arg_update (str): The name of the kwarg for the update_method which must be passed the entry's key.
                    If None, use the same as for the retrieve_method.
                as_json (bool): If true, each field is passed as a json object. If false, all list and map fields are converted to their
                    representation via json.dumps before passing. For a ListResourceDocumentEditor, this only affects the root object.
        """
        self.provider = Common.Session.service_provider(provider)
        self.provider_name = provider
        self.retrieve_method = retrieve_method
        if retrieve_method is not None:
            self.retrieve = getattr(self.provider, retrieve_method)
        if update_method is not None:
            self.update = getattr(self.provider, update_method)
        self.retrieve_path = retrieve_path
        self.retrieve_entry_name_arg = entry_name_arg
        self.update_document_arg = update_document_arg
        self.update_entry_name_arg = (
            entry_name_arg_update
            if entry_name_arg_update is not None
            else entry_name_arg
        )
        self.as_json = as_json
        self.update_method = update_method
        self.message = message

    def retrieve_content(self, selection):
        r_kwargs = {self.retrieve_entry_name_arg: self.selection[self.entry_key]}
        response = self.retrieve(**r_kwargs)
        return self.display_content(response)

    def display_content(self, response):
        return json.dumps(
            Common.Session.jq(self.retrieve_path)
            .input(text=json.dumps(response, default=datetime_hack))
            .first(),
            sort_keys=True,
            indent=2,
        )

    def update_content(self, selection, newcontent):
        try:
            u_kwargs = {
                self.update_entry_name_arg: self.selection[self.entry_key],
                self.update_document_arg: json.loads(newcontent)
                if self.as_json
                else newcontent,
            }
            self.update(**u_kwargs)
            Common.success(
                self.message,
                self.update_method,
                self.provider_name.capitalize(),
                api_kwargs=u_kwargs,
                api_provider=self.provider_name,
                api_method=self.update_method,
            )
        except json.JSONDecodeError as e:
            Common.error(
                "JSON decode error: {0}".format(str(e)),
                self.update_method,
                self.provider_name.capitalize(),
                raw=newcontent,
            )
        except botoerror.ClientError as e:
            Common.clienterror(
                e,
                self.update_method,
                self.provider.capitalize(),
                api_kwargs=u_kwargs,
                api_provider=self.provider_name,
                api_method=self.update_method,
            )
        except botoerror.ParamValidationError as e:
            Common.error(
                "Parameter validation error: {0}".format(str(e)),
                self.update_method,
                self.provider.capitalize(),
                api_kwargs=u_kwargs,
                api_provider=self.provider_name,
                api_method=self.update_method,
            )

    def edit(self, selection):
        content = self.retrieve_content(selection)
        newcontent = Common.Session.textedit(content).strip(" \n\t")
        if content == newcontent:
            Common.Session.set_message("Input unchanged.", Common.color("message_info"))
        self.update_content(selection, newcontent)


class ListResourceDocumentCreator(ListResourceDocumentEditor):
    def __init__(
        self,
        provider,
        create_method,
        create_document_arg,
        initial_document=None,
        as_json=True,
        static_fields={},
        message="Create successful",
    ):
        """
        Initializes a ListResourceDocumentCreator.

            Parameters:
                provider (str): boto3 client provider name
                create_method (str): The method name for the provider to call to create the resource
                create_document_arg (str): The name of the kwarg for the create_method which receives the updated document.
                    If empty, the root level object is passed as a set of kwargs.
                initial_document (dict): The initial document to present for editing. Consider it a template of fields to be filled.
                    Expected to be a dict or dict-like structure that can translate to a json string via json.dumps.
        """
        super().__init__(
            provider,
            None,
            ".",
            create_method,
            None,
            create_document_arg,
            entry_key=None,
            entry_name_arg_update=None,
            as_json=as_json,
        )
        self.initial_document = initial_document
        self.golden = self.display_content(self.initial_document)
        self.static_fields = static_fields
        self.create_method = create_method
        self.provider_name = provider
        self.message = message

    def retrieve_content(self, selection):
        from copy import deepcopy

        return self.display_content(deepcopy(self.initial_document))

    def update_content(self, selection, newcontent):
        if newcontent == self.golden:
            Common.Session.set_message("Cancelled", Common.color("message_info"))
            return
        try:
            update_data = json.loads(newcontent)
        except json.JSONDecodeError as e:
            Common.Session.set_message(
                "JSON parse error: {0}".format(e), Common.color("message_error")
            )
            return
        if self.update_document_arg is not None:
            u_kwargs = {
                self.update_document_arg: update_data if self.as_json else newcontent
            }
        else:
            u_kwargs = {}
            for field in update_data.keys():
                if (
                    not isdumpable(update_data[field])
                    or self.as_json is True
                    or (isinstance(self.as_json, list) and field in self.as_json)
                ):
                    u_kwargs[field] = update_data[field]
                else:
                    u_kwargs[field] = json.dumps(
                        update_data[field], default=datetime_hack
                    )
        for field in self.static_fields.keys():
            u_kwargs[field] = self.static_fields[field]
        try:
            self.update(**u_kwargs)
            Common.success(
                self.message,
                self.create_method,
                self.provider_name.capitalize(),
                api_kwargs=u_kwargs,
                api_provider=self.provider_name,
                api_method=self.create_method,
            )
        except botoerror.ClientError as e:
            Common.clienterror(
                e,
                self.create_method,
                self.provider_name.capitalize(),
                api_kwargs=u_kwargs,
                api_provider=self.provider_name,
                api_method=self.create_method,
            )
        except botoerror.ParamValidationError as e:
            Common.error(
                "Parameter validation error: {0}".format(str(e)),
                self.create_method,
                self.provider_name.capitalize(),
                api_kwargs=u_kwargs,
                api_provider=self.provider_name,
                api_method=self.create_method,
            )

    def edit(self, selection=None):
        super().edit(selection)


class ListResourceFieldsEditor(ListResourceDocumentEditor):
    def __init__(
        self,
        provider,
        retrieve_method,
        retrieve_path,
        update_method,
        entry_name_arg,
        fields,
        entry_key="name",
        entry_name_arg_update=None,
        as_json=True,
    ):
        super().__init__(
            provider,
            retrieve_method,
            retrieve_path,
            update_method,
            entry_name_arg,
            None,
            entry_key="name",
            entry_name_arg_update=entry_name_arg_update,
            as_json=as_json,
        )

        self.fields = fields

    def display_content(self, response):
        content = json.dumps(response, default=datetime_hack)
        ret = {}
        for field, path in self.fields.items():
            ret[field] = Common.Session.jq(path).input(text=content).first()
        return json.dumps(ret, sort_keys=True, indent=2)

    def update_content(self, selection, newcontent):
        update_data = json.loads(newcontent)
        u_kwargs = {self.update_entry_name_arg: self.selection[self.entry_key]}
        for field in self.fields.keys():
            if (
                not isdumpable(update_data[field])
                or self.as_json is True
                or (isinstance(self.as_json, list) and field in self.as_json)
            ):
                u_kwargs[field] = update_data[field]
            else:
                u_kwargs[field] = json.dumps(update_data[field], default=datetime_hack)
        try:
            self.update(**u_kwargs)
            Common.Session.set_message(
                "Update successful", Common.color("message_success")
            )
        except botoerror.ClientError as e:
            Common.Session.set_message(
                "AWS API call error: {0}".format(str(e)), Common.color("message_error")
            )
        except botoerror.ParamValidationError as e:
            Common.Session.set_message(
                "Validation error: {0}".format(str(e)), Common.color("message_error")
            )
